# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The schemas for serializing/deserializing Artifact resources."""
from marshmallow import Schema, fields

from dioptra.restapi.v1.schemas import (
    BasePageSchema,
    GroupIdQueryParametersSchema,
    PagingQueryParametersSchema,
    SearchQueryParametersSchema,
    SortByGetQueryParametersSchema,
    generate_base_resource_ref_schema,
    generate_base_resource_schema,
)

ArtifactRefBaseSchema = generate_base_resource_ref_schema("Artifact")


class ArtifactRefSchema(ArtifactRefBaseSchema):  # type: ignore
    """The reference schema for the data stored in a Artifact resource."""

    artifactUri = fields.URL(
        attribute="artifact_uri",
        metadata=dict(
            description="URL pointing to the location of the Artifact resource."
        ),
        relative=True,
    )


class ArtifactMutableFieldsSchema(Schema):
    """The fields schema for the mutable data in a Artifact resource."""

    description = fields.String(
        attribute="description",
        metadata=dict(description="Description of the Artifact resource."),
        load_default=None,
    )


ArtifactBaseSchema = generate_base_resource_schema("Artifact", snapshot=True)


class ArtifactSchema(ArtifactMutableFieldsSchema, ArtifactBaseSchema):  # type: ignore
    """The schema for the data stored in an Artifact resource."""

    jobId = fields.Int(
        attribute="job_id",
        data_key="job",
        metadata=dict(description="id of the job that produced this Artifact"),
        required=True,
    )
    uri = fields.String(
        attribute="uri",
        metadata=dict(description="URL pointing to the location of the Artifact."),
        required=True,
    )


class ArtifactPageSchema(BasePageSchema):
    """The paged schema for the data stored in an Artifact resource."""

    data = fields.Nested(
        ArtifactSchema,
        many=True,
        metadata=dict(description="List of Artifact resources in the current page."),
    )


class ArtifactGetQueryParameters(
    PagingQueryParametersSchema,
    GroupIdQueryParametersSchema,
    SearchQueryParametersSchema,
    SortByGetQueryParametersSchema,
):
    """The query parameters for the GET method of the /artifacts endpoint."""



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import datetime
from typing import TYPE_CHECKING

from sqlalchemy import ForeignKey, select
from sqlalchemy.orm import Mapped, column_property, mapped_column, relationship

from dioptra.restapi.db.db import bigint, datetimetz, db, intpk, text_

from .constants import group_lock_types
from .locks import GroupLock

if TYPE_CHECKING:
    from .resources import DraftResource, Resource, SharedResource
    from .tags import Tag
    from .users import User

# -- ORM Classes -----------------------------------------------------------------------


class Group(db.Model):  # type: ignore[name-defined]
    __tablename__ = "groups"

    # Database fields
    group_id: Mapped[intpk] = mapped_column(init=False)
    name: Mapped[text_] = mapped_column(nullable=False)
    user_id: Mapped[bigint] = mapped_column(
        ForeignKey("users.user_id"), init=False, nullable=False, index=True
    )
    created_on: Mapped[datetimetz] = mapped_column(init=False, nullable=False)
    last_modified_on: Mapped[datetimetz] = mapped_column(init=False, nullable=False)

    # Derived fields (read-only)
    is_deleted: Mapped[bool] = column_property(
        select(GroupLock.group_id)
        .where(
            GroupLock.group_id == group_id,
            GroupLock.group_lock_type == group_lock_types.DELETE,
        )
        .correlate_except(GroupLock)
        .exists()
    )

    # Relationships
    creator: Mapped["User"] = relationship(back_populates="created_groups")
    members: Mapped[list["GroupMember"]] = relationship(
        init=False, back_populates="group"
    )
    managers: Mapped[list["GroupManager"]] = relationship(
        init=False, back_populates="group"
    )
    draft_resources: Mapped[list["DraftResource"]] = relationship(
        init=False, viewonly=True
    )
    locks: Mapped[list["GroupLock"]] = relationship(init=False, back_populates="group")
    resources: Mapped[list["Resource"]] = relationship(init=False, viewonly=True)
    received_shares: Mapped[list["SharedResource"]] = relationship(
        init=False, viewonly=True
    )
    available_tags: Mapped[list["Tag"]] = relationship(init=False, viewonly=True)

    # Initialize default values using dataclass __post_init__ method
    # https://docs.python.org/3/library/dataclasses.html#dataclasses.__post_init__
    def __post_init__(self) -> None:
        timestamp = datetime.datetime.now(tz=datetime.timezone.utc)
        self.created_on = timestamp
        self.last_modified_on = timestamp


class GroupMember(db.Model):  # type: ignore[name-defined]
    __tablename__ = "group_members"

    # Database fields
    user_id: Mapped[intpk] = mapped_column(ForeignKey("users.user_id"), init=False)
    group_id: Mapped[intpk] = mapped_column(ForeignKey("groups.group_id"), init=False)
    read: Mapped[bool]
    write: Mapped[bool]
    share_read: Mapped[bool]
    share_write: Mapped[bool]

    # Relationships
    user: Mapped["User"] = relationship(
        back_populates="group_memberships", lazy="joined"
    )
    group: Mapped["Group"] = relationship(init=False, back_populates="members")


class GroupManager(db.Model):  # type: ignore[name-defined]
    __tablename__ = "group_managers"

    # Database fields
    user_id: Mapped[intpk] = mapped_column(ForeignKey("users.user_id"), init=False)
    group_id: Mapped[intpk] = mapped_column(ForeignKey("groups.group_id"), init=False)
    owner: Mapped[bool]
    admin: Mapped[bool]

    # Relationships
    user: Mapped["User"] = relationship(
        back_populates="group_managerships", lazy="joined"
    )
    group: Mapped["Group"] = relationship(init=False, back_populates="managers")



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The module defining the endpoints for User resources."""
import uuid
from typing import Any, cast
from urllib.parse import unquote

import structlog
from flask import request
from flask_accepts import accepts, responds
from flask_login import login_required
from flask_restx import Namespace, Resource
from injector import inject
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import models
from dioptra.restapi.v1 import utils
from dioptra.restapi.v1.schemas import IdStatusResponseSchema

from .schema import (
    UserCurrentSchema,
    UserDeleteSchema,
    UserGetQueryParameters,
    UserMutableFieldsSchema,
    UserPageSchema,
    UserPasswordSchema,
    UserSchema,
)
from .service import UserCurrentService, UserIdService, UserService

LOGGER: BoundLogger = structlog.stdlib.get_logger()

api: Namespace = Namespace("Users", description="Users endpoint")


@api.route("/")
class UserEndpoint(Resource):
    @inject
    def __init__(self, user_service: UserService, *args, **kwargs) -> None:
        """Initialize the user resource.

        All arguments are provided via dependency injection.

        Args:
            user_service: A UserService object.
        """
        self._user_service = user_service
        super().__init__(*args, **kwargs)

    @login_required
    @accepts(query_params_schema=UserGetQueryParameters, api=api)
    @responds(schema=UserPageSchema, api=api)
    def get(self):
        """Gets a list of all Users."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="User", request_type="GET"
        )
        parsed_query_params = request.parsed_query_params  # noqa: F841

        search_string = unquote(parsed_query_params["search"])
        page_index = parsed_query_params["index"]
        page_length = parsed_query_params["page_length"]

        users, total_num_users = self._user_service.get(
            search_string=search_string,
            page_index=page_index,
            page_length=page_length,
            log=log,
        )
        return utils.build_paging_envelope(
            "users",
            build_fn=utils.build_user,
            data=users,
            query=search_string,
            draft_type=None,
            group_id=None,
            index=page_index,
            length=page_length,
            total_num_elements=total_num_users,
        )

    @accepts(schema=UserSchema, api=api)
    @responds(schema=UserCurrentSchema, api=api)
    def post(self):
        """Creates a User."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="User", request_type="POST"
        )
        parsed_obj = request.parsed_obj  # noqa: F841

        user = self._user_service.create(
            username=str(parsed_obj["username"]),
            email_address=str(parsed_obj["email"]),
            password=str(parsed_obj["password"]),
            confirm_password=str(parsed_obj["confirm_password"]),
            log=log,
        )
        return utils.build_current_user(user)


@api.route("/<int:id>")
@api.param("id", "ID for the User resource.")
class UserIdEndpoint(Resource):
    @inject
    def __init__(self, user_id_service: UserIdService, *args, **kwargs) -> None:
        """Initialize the user id resource.

        All arguments are provided via dependency injection.

        Args:
            user_service: A UserIdService object.
        """
        self._user_id_service = user_id_service
        super().__init__(*args, **kwargs)

    @login_required
    @responds(schema=UserSchema, api=api)
    def get(self, id: int) -> dict[str, Any]:
        """Gets the User with the provided ID."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="User", request_type="GET", id=id
        )
        user = cast(
            models.User, self._user_id_service.get(id, error_if_not_found=True, log=log)
        )
        return utils.build_user(user)


@api.route("/<int:id>/password")
@api.param("id", "ID for the User resource.")
class UserIdPasswordEndpoint(Resource):
    @inject
    def __init__(self, user_id_service: UserIdService, *args, **kwargs) -> None:
        """Initialize the user id password resource.

        All arguments are provided via dependency injection.

        Args:
            user_service: A UserIdService object.
        """
        self._user_id_service = user_id_service
        super().__init__(*args, **kwargs)

    @login_required
    @accepts(schema=UserPasswordSchema, api=api)
    @responds(schema=IdStatusResponseSchema, api=api)
    def post(self, id: int):
        """Updates the User's password with a given ID."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="User", request_type="POST", id=id
        )
        parsed_obj = request.parsed_obj  # type: ignore # noqa: F841
        return self._user_id_service.change_password(
            user_id=id,
            current_password=parsed_obj["old_password"],
            new_password=parsed_obj["new_password"],
            confirm_new_password=parsed_obj["confirm_new_password"],
            log=log,
        )


@api.route("/current")
class UserCurrentEndpoint(Resource):
    @inject
    def __init__(
        self, user_current_service: UserCurrentService, *args, **kwargs
    ) -> None:
        """Initialize the user current resource.

        All arguments are provided via dependency injection.

        Args:
            user_service: A UserCurrentService object.
        """
        self._user_current_service = user_current_service
        super().__init__(*args, **kwargs)

    @login_required
    @responds(schema=UserCurrentSchema, api=api)
    def get(self):
        """Gets the Current User."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="User", request_type="GET"
        )
        user = self._user_current_service.get(log=log)
        return utils.build_current_user(user)

    @login_required
    @accepts(schema=UserDeleteSchema, api=api)
    @responds(schema=IdStatusResponseSchema, api=api)
    def delete(self) -> dict[str, Any]:
        """Deletes a Current User."""
        log: BoundLogger = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="User",
            request_type="DELETE",
        )
        parsed_obj = request.parsed_obj  # type: ignore
        return self._user_current_service.delete(parsed_obj["password"], log=log)

    @login_required
    @accepts(schema=UserMutableFieldsSchema, api=api)
    @responds(schema=UserCurrentSchema, api=api)
    def put(self) -> dict[str, Any]:
        """Modifies the Current User"""
        log: BoundLogger = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="User",
            request_type="PUT",
        )
        parsed_obj = request.parsed_obj  # type: ignore
        user = self._user_current_service.modify(
            username=parsed_obj["username"],
            email_address=parsed_obj["email"],
            error_if_not_found=True,
            log=log,
        )
        return utils.build_current_user(user)


@api.route("/current/password")
class UserCurrentPasswordEndpoint(Resource):
    @inject
    def __init__(
        self, user_current_service: UserCurrentService, *args, **kwargs
    ) -> None:
        """Initialize the user current resource.

        All arguments are provided via dependency injection.

        Args:
            user_service: A UserCurrentService object.
        """
        self._user_current_service = user_current_service
        super().__init__(*args, **kwargs)

    @login_required
    @accepts(schema=UserPasswordSchema, api=api)
    @responds(schema=IdStatusResponseSchema, api=api)
    def post(self):
        """Updates the Current User's password."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="User",
            request_type="POST",
        )
        parsed_obj = request.parsed_obj  # noqa: F841
        return self._user_current_service.change_password(
            current_password=parsed_obj["old_password"],
            new_password=parsed_obj["new_password"],
            confirm_new_password=parsed_obj["confirm_new_password"],
            log=log,
        )



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The module defining the endpoints for Group resources."""
import uuid
from typing import cast
from urllib.parse import unquote

import structlog
from flask import request
from flask_accepts import accepts, responds
from flask_login import login_required
from flask_restx import Namespace, Resource
from injector import inject
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import models
from dioptra.restapi.v1 import utils
from dioptra.restapi.v1.schemas import IdStatusResponseSchema

from .schema import (
    GroupGetQueryParameters,
    GroupMemberMutableFieldsSchema,
    GroupMemberSchema,
    GroupMutableFieldsSchema,
    GroupPageSchema,
    GroupSchema,
)
from .service import GroupIdService, GroupService

LOGGER: BoundLogger = structlog.stdlib.get_logger()

api: Namespace = Namespace("Groups", description="Groups endpoint")


@api.route("/")
class GroupEndpoint(Resource):
    @inject
    def __init__(self, group_service: GroupService, *args, **kwargs) -> None:
        """Initialize the group service

        All arguments are provided via dependency injection.

        Args:
            group_service: A GroupService object.
        """
        self._group_service = group_service
        super().__init__(*args, **kwargs)

    @login_required
    @accepts(query_params_schema=GroupGetQueryParameters, api=api)
    @responds(schema=GroupPageSchema, api=api)
    def get(self):
        """Gets a list of all Group resources."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Group", request_type="GET"
        )
        parsed_query_params = request.parsed_query_params  # noqa: F841

        search_string = unquote(parsed_query_params["search"])
        page_index = parsed_query_params["index"]
        page_length = parsed_query_params["page_length"]

        groups, total_num_groups = self._group_service.get(
            search_string=search_string,
            page_index=page_index,
            page_length=page_length,
            log=log,
        )
        return utils.build_paging_envelope(
            "groups",
            build_fn=utils.build_group,
            data=groups,
            group_id=None,
            query=search_string,
            draft_type=None,
            index=page_index,
            length=page_length,
            total_num_elements=total_num_groups,
        )

    @login_required
    @accepts(schema=GroupSchema, api=api)
    @responds(schema=GroupSchema, api=api)
    def _post(self):
        """Creates a Group resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Group", request_type="POST"
        )
        parsed_obj = request.parsed_obj  # noqa: F841

        group = self._group_service.create(name=parsed_obj["name"], log=log)
        return utils.build_group(group)


@api.route("/<int:id>")
@api.param("id", "ID for the Group resource.")
class GroupIdEndpoint(Resource):
    @inject
    def __init__(self, group_id_service: GroupIdService, *args, **kwargs) -> None:
        """Initialize the group service

        All arguments are provided via dependency injection.

        Args:
            group_service: A GroupService object.
        """
        self._group_id_service = group_id_service
        super().__init__(*args, **kwargs)

    @login_required
    @responds(schema=GroupSchema, api=api)
    def get(self, id: int):
        """Gets a Group resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Group", request_type="GET", id=id
        )
        group = cast(
            models.Group,
            self._group_id_service.get(id, error_if_not_found=True, log=log),
        )
        return utils.build_group(group)

    @login_required
    @responds(schema=IdStatusResponseSchema, api=api)
    def _delete(self, id: int):
        """Deletes a Group resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Group", request_type="DELETE", id=id
        )
        return self._group_id_service.delete(id, log=log)

    @login_required
    @accepts(schema=GroupMutableFieldsSchema, api=api)
    @responds(schema=GroupSchema, api=api)
    def _put(self, id: int):
        """Modifies a Group resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Group", request_type="PUT", id=id
        )
        parsed_obj = request.parsed_obj  # type: ignore # noqa: F841
        group = cast(
            models.Group,
            self._group_id_service.modify(
                id, name=parsed_obj["name"], error_if_not_found=True, log=log
            ),
        )
        return utils.build_group(group)


# @api.route("/<int:id>/members")
@api.param("id", "ID for the Group resource.")
class GroupIdMembersEndpoint(Resource):
    @login_required
    @responds(schema=GroupMemberSchema(many=True), api=api)
    def get(self, id: int):
        """Gets a list of Member's Group Permissions."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="GroupMember",
            request_type="GET",
            id=id,
        )
        log.debug("Request received")

    @login_required
    @accepts(schema=GroupMemberSchema, api=api)
    @responds(schema=GroupMemberSchema, api=api)
    def post(self, id: int):
        """Adds a Member to the Group with default permissions."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="GroupMember",
            request_type="POST",
            id=id,
        )
        log.debug("Request received")
        parsed_obj = request.parsed_obj  # type: ignore # noqa: F841

    @login_required
    @accepts(schema=GroupMemberSchema(many=True), api=api)
    @responds(schema=GroupMemberSchema(many=True), api=api)
    def put(self, id: int):
        """Modifies all Group Members' permissions."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="GroupMember",
            request_type="POST",
            id=id,
        )
        log.debug("Request received")
        parsed_obj = request.parsed_obj  # type: ignore # noqa: F841


# @api.route("/<int:id>/members/<int:user_id>")
@api.param("id", "ID for the Group resource.")
@api.param("user_id", "ID for the User resource.")
class GroupIdMembersUserIdEndpoint(Resource):
    @login_required
    @responds(schema=GroupMemberSchema, api=api)
    def get(self, id: int, user_id: int):
        """Gets a Member's Group permissions."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="GroupMember",
            request_type="GET",
            id=id,
        )
        log.debug("Request received")

    @login_required
    @responds(schema=IdStatusResponseSchema, api=api)
    def delete(self, id: int, user_id: int):
        """Removes a Member from the Group."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="GroupMember",
            request_type="DELETE",
            id=id,
        )
        log.debug("Request received")

    @login_required
    @accepts(schema=GroupMemberMutableFieldsSchema, api=api)
    @responds(schema=GroupMemberSchema, api=api)
    def patch(self, id: int, user_id: int):
        """Modifies a Member's Group permissions."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="GroupMember",
            request_type="PATCH",
            id=id,
        )
        log.debug("Request received")
        parsed_obj = request.parsed_obj  # type: ignore # noqa: F841



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""Test suite for using the v1 ORM models to insert data in the database.

These tests exercise all of the v1 tables to ensure that the ORM models are configured
correctly and respect the data integrity constraints defined in the database schema.
These tests are a temporary measure that should eventually be replaced by integration
tests for the REST API.
"""
import uuid

import pytest
from faker import Faker
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy.exc import IntegrityError

from dioptra.restapi.db import models
from dioptra.restapi.db.models.constants import (
    group_lock_types,
    resource_lock_types,
    user_lock_types,
)

from .lib import db as libdb
from .lib.db import views as viewsdb

# -- Fixtures --------------------------------------------------------------------------


@pytest.fixture
def fake_data(faker: Faker) -> libdb.FakeData:
    return libdb.FakeData(faker)


@pytest.fixture
def account(db: SQLAlchemy, fake_data: libdb.FakeData) -> libdb.FakeAccount:
    new_account = fake_data.account()
    db.session.add(new_account.user)
    db.session.add(new_account.group)
    db.session.commit()
    return new_account


@pytest.fixture
def second_account(
    db: SQLAlchemy, account: libdb.FakeData, fake_data: libdb.FakeData
) -> libdb.FakeAccount:
    new_account = fake_data.account()
    db.session.add(new_account.user)
    db.session.add(new_account.group)
    db.session.commit()
    return new_account


@pytest.fixture
def deleted_user_id(
    db: SQLAlchemy,
    account: libdb.FakeAccount,
) -> int:
    deleted_user_id = account.user.user_id
    deleted_user_lock = models.UserLock(
        user_lock_type=user_lock_types.DELETE,
        user=account.user,
    )
    db.session.add(deleted_user_lock)
    db.session.commit()
    return deleted_user_id


@pytest.fixture
def deleted_group_id(
    db: SQLAlchemy,
    account: libdb.FakeAccount,
) -> int:
    deleted_group_id = account.group.group_id
    deleted_group_lock = models.GroupLock(
        group_lock_type=group_lock_types.DELETE,
        group=account.group,
    )
    db.session.add(deleted_group_lock)
    db.session.commit()
    return deleted_group_id


@pytest.fixture
def tag_id(
    db: SQLAlchemy,
    account: libdb.FakeAccount,
    fake_data: libdb.FakeData,
) -> int:
    new_tag = fake_data.tag(creator=account.user, group=account.group)
    db.session.add(new_tag)
    db.session.commit()
    return new_tag.tag_id


@pytest.fixture
def registered_queue_id(
    db: SQLAlchemy,
    account: libdb.FakeAccount,
    fake_data: libdb.FakeData,
) -> int:
    new_queue = fake_data.queue(creator=account.user, group=account.group)
    db.session.add(new_queue)
    db.session.commit()
    return new_queue.resource_id


@pytest.fixture
def registered_queue_id_with_description(
    db: SQLAlchemy,
    account: libdb.FakeAccount,
    fake_data: libdb.FakeData,
) -> int:
    new_queue = fake_data.queue(
        creator=account.user, group=account.group, include_description=True
    )
    db.session.add(new_queue)
    db.session.commit()
    return new_queue.resource_id


@pytest.fixture
def draft_resource_id_for_new_queue_resource(
    db: SQLAlchemy,
    account: libdb.FakeAccount,
    fake_data: libdb.FakeData,
) -> int:
    new_queue = fake_data.queue(
        creator=account.user, group=account.group, include_description=True
    )
    draft_payload = {
        "name": new_queue.name,
        "description": new_queue.description,
        # If doing an update, this is the ID for snapshot being modified.
        # If None, then a new resource will be created.
        "resource_snapshot_id": None,
        # If doing an update, this is the ID for the resource being modified.
        # If None, then a new resource will be created.
        "resource_id": None,
    }
    draft_resource = models.DraftResource(
        resource_type="queue",
        target_owner=account.group,
        payload=draft_payload,
        creator=account.user,
    )
    db.session.add(draft_resource)
    db.session.commit()
    return draft_resource.draft_resource_id


@pytest.fixture
def edited_queue_id(
    db: SQLAlchemy,
    faker: Faker,
    registered_queue_id: int,
) -> int:
    queue_v1 = viewsdb.get_latest_queue(db, resource_id=registered_queue_id)

    if queue_v1 is None:
        raise ValueError("Failed to retrieve the latest queue for test.")

    queue_v2 = models.Queue(
        name=faker.sentence(nb_words=3, variable_nb_words=True),
        description=queue_v1.description,
        resource=queue_v1.resource,
        creator=queue_v1.creator,
    )
    db.session.add(queue_v2)
    db.session.commit()

    queue_v2 = viewsdb.get_latest_queue(db, resource_id=registered_queue_id)

    if queue_v2 is None:
        raise ValueError("Failed to retrieve the latest queue for test.")

    queue_v3 = models.Queue(
        name=queue_v2.name,
        description=faker.sentence(),
        resource=queue_v2.resource,
        creator=queue_v2.creator,
    )
    db.session.add(queue_v3)
    db.session.commit()

    return registered_queue_id


@pytest.fixture
def deleted_queue_id(
    db: SQLAlchemy,
    registered_queue_id: int,
) -> int:
    queue = viewsdb.get_latest_queue(db, resource_id=registered_queue_id)

    if queue is None:
        raise ValueError("Failed to retrieve the latest queue for test.")

    deleted_resource_lock = models.ResourceLock(
        resource_lock_type=resource_lock_types.DELETE,
        resource=queue.resource,
    )
    db.session.add(deleted_resource_lock)
    db.session.commit()
    return registered_queue_id


@pytest.fixture
def registered_and_tagged_queue_id(
    db: SQLAlchemy,
    registered_queue_id: int,
    tag_id: int,
) -> int:
    tag = db.session.get_one(models.Tag, tag_id)
    queue = viewsdb.get_latest_queue(db, resource_id=registered_queue_id)

    if queue is None:
        raise ValueError("Failed to retrieve the latest queue for test.")

    tag.resources.append(queue.resource)
    db.session.commit()

    return queue.resource_id


@pytest.fixture
def shared_resource_id_for_queue(
    db: SQLAlchemy,
    account: libdb.FakeAccount,
    second_account: libdb.FakeAccount,
    registered_queue_id: int,
) -> int:
    queue = viewsdb.get_latest_queue(db, resource_id=registered_queue_id)

    if queue is None:
        raise ValueError("Failed to retrieve the latest queue for test.")

    # The first user creates a read-only share of a queue resource for the second user's
    # default group.
    shared_resource = models.SharedResource(
        read=True,
        write=False,
        recipient=second_account.group,
        resource=queue.resource,
        creator=account.user,
    )
    db.session.add(shared_resource)
    db.session.commit()

    return shared_resource.shared_resource_id


@pytest.fixture
def registered_experiment_id(
    db: SQLAlchemy,
    account: libdb.FakeAccount,
    fake_data: libdb.FakeData,
) -> int:
    new_experiment = fake_data.experiment(creator=account.user, group=account.group)
    db.session.add(new_experiment)
    db.session.commit()
    return new_experiment.resource_id


@pytest.fixture
def registered_experiment_id_with_description(
    db: SQLAlchemy,
    account: libdb.FakeAccount,
    fake_data: libdb.FakeData,
) -> int:
    new_experiment = fake_data.experiment(
        creator=account.user, group=account.group, include_description=True
    )
    db.session.add(new_experiment)
    db.session.commit()
    return new_experiment.resource_id


@pytest.fixture
def registered_and_tagged_experiment_id(
    db: SQLAlchemy,
    registered_experiment_id: int,
    tag_id: int,
) -> int:
    tag = db.session.get_one(models.Tag, tag_id)
    experiment = viewsdb.get_latest_experiment(db, resource_id=registered_experiment_id)

    if experiment is None:
        raise ValueError("Failed to retrieve the latest queue for test.")

    tag.resources.append(experiment.resource)
    db.session.commit()

    return experiment.resource_id


@pytest.fixture
def registered_str_parameter_type_id(
    db: SQLAlchemy,
    account: libdb.FakeAccount,
    fake_data: libdb.FakeData,
) -> int:
    new_plugin_task_parameter_type = fake_data.plugin_task_parameter_type(
        creator=account.user, group=account.group, name="str"
    )
    db.session.add(new_plugin_task_parameter_type)
    db.session.commit()
    return new_plugin_task_parameter_type.resource_id


@pytest.fixture
def registered_plugin_id(
    db: SQLAlchemy,
    account: libdb.FakeAccount,
    fake_data: libdb.FakeData,
    registered_str_parameter_type_id: int,
) -> int:
    registered_str_parameter_type = viewsdb.get_latest_plugin_task_parameter_type(
        db, resource_id=registered_str_parameter_type_id
    )

    if registered_str_parameter_type is None:
        raise ValueError(
            "Failed to retrieve the latest plugin task parameter type for test."
        )

    new_plugin = fake_data.plugin(
        creator=account.user,
        group=account.group,
        str_parameter_type=registered_str_parameter_type,
    )
    db.session.add(new_plugin.plugin)
    db.session.add(new_plugin.init_plugin_file)
    db.session.add(new_plugin.plugin_file)
    db.session.add(new_plugin.plugin_task)
    db.session.commit()
    return new_plugin.plugin.resource_id


@pytest.fixture
def registered_entry_point_id(
    db: SQLAlchemy,
    account: libdb.FakeAccount,
    fake_data: libdb.FakeData,
    registered_queue_id: int,
    registered_plugin_id: int,
) -> int:
    registered_queue = viewsdb.get_latest_queue(db, resource_id=registered_queue_id)

    if registered_queue is None:
        raise ValueError("Failed to retrieve the latest queue for test.")

    registered_plugin = viewsdb.get_latest_plugin(db, resource_id=registered_plugin_id)

    if registered_plugin is None:
        raise ValueError("Failed to retrieve the latest plugin for test.")

    latest_plugin_files = viewsdb.get_latest_plugin_files(
        db,
        plugin_resource_id=registered_plugin_id,
    )
    new_entry_point = fake_data.entry_point(
        creator=account.user,
        group=account.group,
        plugin=registered_plugin,
        plugin_files=latest_plugin_files,
        queue=registered_queue,
    )
    db.session.add(new_entry_point)
    db.session.commit()

    return new_entry_point.resource_id


@pytest.fixture
def registered_experiment_entry_point_id(
    db: SQLAlchemy,
    registered_experiment_id: int,
    registered_entry_point_id: int,
) -> int:
    registered_experiment = viewsdb.get_latest_experiment(
        db, resource_id=registered_experiment_id
    )

    if registered_experiment is None:
        raise ValueError("Failed to retrieve the latest experiment for test.")

    registered_entry_point = viewsdb.get_latest_entry_point(
        db, resource_id=registered_entry_point_id
    )

    if registered_entry_point is None:
        raise ValueError("Failed to retrieve the latest entry point for test.")

    registered_experiment.children.append(registered_entry_point.resource)
    db.session.commit()

    return registered_experiment.resource_id


@pytest.fixture
def registered_job_id(
    db: SQLAlchemy,
    account: libdb.FakeAccount,
    fake_data: libdb.FakeData,
    registered_experiment_entry_point_id: int,
) -> int:
    registered_experiment = viewsdb.get_latest_experiment(
        db, resource_id=registered_experiment_entry_point_id
    )

    if registered_experiment is None:
        raise ValueError("Failed to retrieve the latest experiment for test.")

    experiment_entry_points = viewsdb.get_latest_experiment_entry_points(
        db, experiment_resource_id=registered_experiment_entry_point_id
    )

    if not experiment_entry_points:
        raise ValueError("No entry points associated with experiment resource.")

    entry_point_queues = viewsdb.get_latest_entry_point_queues(
        db, entry_point_resource_id=experiment_entry_points[0].resource_id
    )

    if not entry_point_queues:
        raise ValueError("No queues associated with entry point resource.")

    new_job = fake_data.job(
        creator=account.user,
        group=account.group,
        entry_point=experiment_entry_points[0],
        experiment=registered_experiment,
        queue=entry_point_queues[0],
    )
    db.session.add(new_job)
    db.session.commit()

    return new_job.resource_id


@pytest.fixture
def registered_child_job_id(
    db: SQLAlchemy,
    account: libdb.FakeAccount,
    fake_data: libdb.FakeData,
    registered_experiment_entry_point_id: int,
    registered_job_id: int,
) -> int:
    registered_experiment = viewsdb.get_latest_experiment(
        db, resource_id=registered_experiment_entry_point_id
    )

    if registered_experiment is None:
        raise ValueError("Failed to retrieve the latest experiment for test.")

    experiment_entry_points = viewsdb.get_latest_experiment_entry_points(
        db, experiment_resource_id=registered_experiment_entry_point_id
    )

    if not experiment_entry_points:
        raise ValueError("No entry points associated with experiment resource.")

    entry_point_queues = viewsdb.get_latest_entry_point_queues(
        db, entry_point_resource_id=experiment_entry_points[0].resource_id
    )

    if not entry_point_queues:
        raise ValueError("No queues associated with entry point resource.")

    parent_job = viewsdb.get_latest_job(db, resource_id=registered_job_id)

    if parent_job is None:
        raise ValueError("Failed to retrieve the parent job for test.")

    new_child_job = fake_data.job(
        creator=account.user,
        group=account.group,
        entry_point=experiment_entry_points[0],
        experiment=registered_experiment,
        queue=entry_point_queues[0],
    )

    # Make new_child_job depend on parent_job, meaning it should not start until
    # parent_job ends.
    new_child_job.parents.append(parent_job.resource)

    db.session.add(new_child_job)
    db.session.commit()

    return new_child_job.resource_id


@pytest.fixture
def registered_mlflow_run_job_resource_id(
    db: SQLAlchemy,
    account: libdb.FakeAccount,
    fake_data: libdb.FakeData,
    registered_job_id: int,
) -> int:
    job = viewsdb.get_latest_job(db, resource_id=registered_job_id)

    if job is None:
        raise ValueError("Failed to retrieve the latest job for test.")

    job_mlflow_run = fake_data.job_mlflow_run(job=job)
    db.session.add(job_mlflow_run)
    db.session.commit()

    return job_mlflow_run.job_resource_id


@pytest.fixture
def registered_artifact_id(
    db: SQLAlchemy,
    account: libdb.FakeAccount,
    fake_data: libdb.FakeData,
    registered_job_id: int,
) -> int:
    job = viewsdb.get_latest_job(db, resource_id=registered_job_id)

    if job is None:
        raise ValueError("Failed to retrieve the latest job for test.")

    artifact = fake_data.artifact(creator=account.user, group=account.group, job=job)
    db.session.add(artifact)
    db.session.commit()

    return artifact.resource_id


@pytest.fixture
def registered_ml_model_id(
    db: SQLAlchemy,
    account: libdb.FakeAccount,
    fake_data: libdb.FakeData,
    registered_artifact_id: int,
) -> int:
    artifact = viewsdb.get_latest_artifact(db, resource_id=registered_artifact_id)

    if artifact is None:
        raise ValueError("Failed to retrieve the latest artifact for test.")

    ml_model = fake_data.ml_model(
        creator=account.user, group=account.group, artifact=artifact
    )
    db.session.add(ml_model)
    db.session.commit()

    return ml_model.resource_id


# -- Tests -----------------------------------------------------------------------------


def test_db_user_not_deleted(db: SQLAlchemy, account: libdb.FakeAccount) -> None:
    user = db.session.get_one(models.User, account.user.user_id)
    delete_lock = [x for x in user.locks if x.user_lock_type == user_lock_types.DELETE]
    assert len(delete_lock) == 0
    assert isinstance(user.is_deleted, bool)
    assert not user.is_deleted


def test_db_delete_user(db: SQLAlchemy, deleted_user_id: int) -> None:
    deleted_user = db.session.get_one(models.User, deleted_user_id)
    delete_lock = [
        x for x in deleted_user.locks if x.user_lock_type == user_lock_types.DELETE
    ]
    assert len(delete_lock) == 1
    assert isinstance(deleted_user.is_deleted, bool)
    assert deleted_user.is_deleted


def test_db_group_not_deleted(db: SQLAlchemy, account: libdb.FakeAccount) -> None:
    group = db.session.get_one(models.Group, account.group.group_id)
    delete_lock = [
        x for x in group.locks if x.group_lock_type == group_lock_types.DELETE
    ]
    assert len(delete_lock) == 0
    assert isinstance(group.is_deleted, bool)
    assert not group.is_deleted


def test_db_delete_group(db: SQLAlchemy, deleted_group_id: int) -> None:
    deleted_group = db.session.get_one(models.Group, deleted_group_id)
    delete_lock = [
        x for x in deleted_group.locks if x.group_lock_type == group_lock_types.DELETE
    ]
    assert len(delete_lock) == 1
    assert isinstance(deleted_group.is_deleted, bool)
    assert deleted_group.is_deleted


def test_db_add_queue(
    db: SQLAlchemy, account: libdb.FakeAccount, registered_queue_id: int
) -> None:
    new_queue = viewsdb.get_latest_queue(db, resource_id=registered_queue_id)

    assert new_queue is not None
    assert new_queue.creator.user_id == account.user.user_id
    assert new_queue.resource.owner.group_id == account.group.group_id
    assert isinstance(new_queue.name, str)
    assert new_queue.description is None


def test_db_add_queue_with_description(
    db: SQLAlchemy,
    account: libdb.FakeAccount,
    registered_queue_id_with_description: int,
) -> None:
    new_queue = viewsdb.get_latest_queue(
        db, resource_id=registered_queue_id_with_description
    )

    assert new_queue is not None
    assert new_queue.creator.user_id == account.user.user_id
    assert new_queue.resource.owner.group_id == account.group.group_id
    assert isinstance(new_queue.name, str)
    assert new_queue.description is not None and isinstance(new_queue.description, str)


def test_db_add_queue_using_data_in_a_draft_resource(
    db: SQLAlchemy,
    account: libdb.FakeAccount,
    draft_resource_id_for_new_queue_resource: int,
) -> None:
    draft_resource = db.session.get_one(
        models.DraftResource, draft_resource_id_for_new_queue_resource
    )
    created_draft_resource_ids = {
        x.draft_resource_id for x in account.user.created_drafts
    }
    assert draft_resource_id_for_new_queue_resource in created_draft_resource_ids

    # Note: What follows assumes that we know the value of draft_resource.resource_type
    # ahead of time. In practice this will not be the case, and the appropriate ORM
    # model class will need to be retrieved based on the value of
    # draft_resource.resource_type.
    new_queue = models.Queue(
        name=draft_resource.payload["name"],
        description=draft_resource.payload["description"],
        resource=models.Resource(
            resource_type=draft_resource.resource_type,
            owner=draft_resource.target_owner,
        ),
        creator=draft_resource.creator,
    )
    db.session.add(new_queue)
    db.session.delete(draft_resource)
    db.session.commit()

    deleted_draft_resource = db.session.get(
        models.DraftResource, draft_resource_id_for_new_queue_resource
    )

    assert new_queue.resource_snapshot_id is not None
    assert new_queue.resource_id is not None
    assert new_queue.creator.user_id == account.user.user_id
    assert new_queue.resource.owner.group_id == account.group.group_id
    assert isinstance(new_queue.name, str)
    assert new_queue.description is not None and isinstance(new_queue.description, str)
    assert deleted_draft_resource is None


def test_db_share_queue_resource(
    db: SQLAlchemy,
    account: libdb.FakeAccount,
    second_account: libdb.FakeAccount,
    shared_resource_id_for_queue: int,
) -> None:
    shared_resource = db.session.get_one(
        models.SharedResource, shared_resource_id_for_queue
    )
    received_shared_resource_ids = {
        x.shared_resource_id for x in second_account.group.received_shares
    }
    created_shared_resource_ids = {
        x.shared_resource_id for x in account.user.created_shares
    }
    shared_queue = viewsdb.get_latest_queue(db, resource_id=shared_resource.resource_id)

    assert shared_resource.shared_resource_id in received_shared_resource_ids
    assert shared_resource.shared_resource_id in created_shared_resource_ids
    assert shared_queue is not None


def test_db_queue_resource_not_deleted(
    db: SQLAlchemy, registered_queue_id: int
) -> None:
    queue = viewsdb.get_latest_queue(db, resource_id=registered_queue_id)
    assert queue is not None

    queue_resource = db.session.get_one(models.Resource, registered_queue_id)
    assert isinstance(queue_resource.is_deleted, bool)
    assert not queue_resource.is_deleted


def test_db_delete_queue_resource(db: SQLAlchemy, deleted_queue_id: int) -> None:
    deleted_queue = viewsdb.get_latest_queue(db, resource_id=deleted_queue_id)
    assert deleted_queue is None

    deleted_queue_resource = db.session.get_one(models.Resource, deleted_queue_id)
    assert isinstance(deleted_queue_resource.is_deleted, bool)
    assert deleted_queue_resource.is_deleted


def test_db_queue_last_modified_on(db: SQLAlchemy, edited_queue_id: int) -> None:
    queue = viewsdb.get_latest_queue(db, resource_id=edited_queue_id)

    assert queue is not None
    assert len(queue.resource.versions) > 1
    assert queue.resource.last_modified_on > queue.resource.created_on
    assert queue.created_on == queue.resource.last_modified_on


def test_db_add_experiment(
    db: SQLAlchemy, account: libdb.FakeAccount, registered_experiment_id: int
) -> None:
    new_experiment = viewsdb.get_latest_experiment(
        db, resource_id=registered_experiment_id
    )

    assert new_experiment is not None
    assert new_experiment.creator.user_id == account.user.user_id
    assert new_experiment.resource.owner.group_id == account.group.group_id
    assert isinstance(new_experiment.name, str)
    assert new_experiment.description is None


def test_db_add_experiment_with_description(
    db: SQLAlchemy,
    account: libdb.FakeAccount,
    registered_experiment_id_with_description: int,
) -> None:
    new_experiment = viewsdb.get_latest_experiment(
        db, resource_id=registered_experiment_id_with_description
    )

    assert new_experiment is not None
    assert new_experiment.creator.user_id == account.user.user_id
    assert new_experiment.resource.owner.group_id == account.group.group_id
    assert isinstance(new_experiment.name, str)
    assert new_experiment.description is not None and isinstance(
        new_experiment.description, str
    )


def test_db_tag_resources(
    db: SQLAlchemy,
    account: libdb.FakeAccount,
    registered_and_tagged_queue_id: int,
    registered_and_tagged_experiment_id: int,
    tag_id: int,
) -> None:
    queue = viewsdb.get_latest_queue(db, resource_id=registered_and_tagged_queue_id)
    experiment = viewsdb.get_latest_experiment(
        db, resource_id=registered_and_tagged_experiment_id
    )

    assert queue is not None and experiment is not None
    assert tag_id in {x.tag_id for x in queue.tags}
    assert tag_id in {x.tag_id for x in experiment.tags}
    assert tag_id in {x.tag_id for x in account.user.created_tags}
    assert tag_id in {x.tag_id for x in account.group.available_tags}


def test_db_add_plugin_task_parameter_type(
    db: SQLAlchemy, account: libdb.FakeAccount, registered_str_parameter_type_id: int
) -> None:
    new_plugin_task_parameter_type = viewsdb.get_latest_plugin_task_parameter_type(
        db, resource_id=registered_str_parameter_type_id
    )

    assert new_plugin_task_parameter_type is not None
    assert new_plugin_task_parameter_type.creator.user_id == account.user.user_id
    assert (
        new_plugin_task_parameter_type.resource.owner.group_id == account.group.group_id
    )
    assert isinstance(new_plugin_task_parameter_type.name, str)


def test_db_add_plugin(
    db: SQLAlchemy, account: libdb.FakeAccount, registered_plugin_id: int
) -> None:
    new_plugin = viewsdb.get_latest_plugin(db, resource_id=registered_plugin_id)
    new_plugin_files = viewsdb.get_latest_plugin_files(
        db, plugin_resource_id=registered_plugin_id
    )

    assert new_plugin is not None
    assert new_plugin.creator.user_id == account.user.user_id
    assert new_plugin.resource.owner.group_id == account.group.group_id
    assert len(new_plugin_files) > 0

    for plugin_file in new_plugin_files:
        assert plugin_file.creator.user_id == account.user.user_id
        assert plugin_file.resource.owner.group_id == account.group.group_id


def test_db_add_entry_point(
    db: SQLAlchemy, account: libdb.FakeAccount, registered_entry_point_id: int
) -> None:
    new_entry_point = viewsdb.get_latest_entry_point(
        db, resource_id=registered_entry_point_id
    )

    assert new_entry_point is not None
    assert new_entry_point.creator.user_id == account.user.user_id
    assert new_entry_point.resource.owner.group_id == account.group.group_id
    assert len(new_entry_point.entry_point_plugin_files) > 0


def test_db_add_job(
    db: SQLAlchemy,
    account: libdb.FakeAccount,
    registered_job_id: int,
) -> None:
    new_job = viewsdb.get_latest_job(db, resource_id=registered_job_id)

    assert new_job is not None
    assert new_job.creator.user_id == account.user.user_id
    assert new_job.resource.owner.group_id == account.group.group_id
    assert new_job.entry_point_job.entry_point is not None
    assert new_job.experiment_job.experiment is not None
    assert new_job.queue_job.queue is not None
    assert new_job.mlflow_run is None


def test_db_add_job_that_depends_on_another_job(
    db: SQLAlchemy,
    registered_job_id: int,
    registered_child_job_id: int,
) -> None:
    child_job = viewsdb.get_latest_job(db, resource_id=registered_child_job_id)

    assert child_job is not None
    assert registered_job_id in {x.resource_id for x in child_job.parents}


def test_db_add_mlflow_run_to_job(
    db: SQLAlchemy,
    registered_mlflow_run_job_resource_id: int,
) -> None:
    job = viewsdb.get_latest_job(db, resource_id=registered_mlflow_run_job_resource_id)

    assert job is not None
    assert job.mlflow_run is not None
    assert isinstance(job.mlflow_run.mlflow_run_id, uuid.UUID)


@pytest.mark.skip(reason="Changes to the MlModel ORM object make this test incorrect.")
def test_db_add_artifact_to_job(
    db: SQLAlchemy,
    account: libdb.FakeAccount,
    registered_job_id: int,
    registered_artifact_id: int,
) -> None:
    new_artifact = viewsdb.get_latest_artifact(db, resource_id=registered_artifact_id)

    # Validate that the job/artifact association worked as expected, should be identical
    # to new_artifact
    job_artifacts = viewsdb.get_latest_job_artifacts(
        db, job_resource_id=registered_job_id
    )

    assert new_artifact is not None
    assert len(job_artifacts) == 1
    assert new_artifact.resource_snapshot_id == job_artifacts[0].resource_snapshot_id
    assert new_artifact.creator.user_id == account.user.user_id
    assert new_artifact.resource.owner.group_id == account.group.group_id
    assert new_artifact.uri.startswith("s3://")


@pytest.mark.skip(reason="Changes to the MlModel ORM object make this test incorrect.")
def test_db_add_ml_model_to_job(
    db: SQLAlchemy,
    account: libdb.FakeAccount,
    registered_ml_model_id: int,
) -> None:
    new_ml_model = viewsdb.get_latest_ml_model(db, resource_id=registered_ml_model_id)

    assert new_ml_model is not None
    assert new_ml_model.creator.user_id == account.user.user_id
    assert new_ml_model.resource.owner.group_id == account.group.group_id
    assert new_ml_model.artifact.creator.user_id == account.user.user_id
    assert new_ml_model.artifact.resource.owner.group_id == account.group.group_id
    assert isinstance(new_ml_model.name, str)


@pytest.mark.skip(reason="Changes to the MlModel ORM object make this test incorrect.")
def test_db_invalid_resource_dependency_fails(
    db: SQLAlchemy,
    registered_plugin_id: int,
    registered_ml_model_id: int,
) -> None:
    plugin = viewsdb.get_latest_plugin(db, resource_id=registered_plugin_id)
    ml_model = viewsdb.get_latest_ml_model(db, resource_id=registered_ml_model_id)

    assert plugin is not None and ml_model is not None

    with pytest.raises(IntegrityError):
        try:
            ml_model.parents.append(plugin.resource)
            db.session.commit()

        except IntegrityError as err:
            db.session.rollback()
            raise err



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import enum


class IssueSeverity(enum.Enum):
    """
    Validation issue severity levels
    """

    ERROR = enum.auto()
    WARNING = enum.auto()


class IssueType(enum.Enum):
    """
    Validation issue types
    """

    SYNTAX = enum.auto()
    SCHEMA = enum.auto()
    SEMANTIC = enum.auto()
    TYPE = enum.auto()


class ValidationIssue:
    """
    Represents a validation "issue", including errors, warnings, etc.
    """

    def __init__(self, type_: IssueType, severity: IssueSeverity, message: str) -> None:
        self.type = type_
        self.severity = severity
        self.message = message

    def __repr__(self) -> str:
        value = "ValidationIssue({!s}, {!s}, {!r})".format(
            self.type, self.severity, self.message
        )

        return value

    def __str__(self) -> str:
        value = "{}.{}: {}".format(
            self.type.name.lower(), self.severity.name.lower(), self.message
        )

        return value



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
#
# This is an adaptation of the work
# https://github.com/ashep29/confluence/blob/748c71e848b8a397df6ab0d6173ba890d8b585e2/confluence.py  # noqa: B950
# See copyright below.
#
# Copyright (c) 2021 ashep29
# Distributed under the terms of the MIT License.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

from collections import defaultdict
from typing import cast

import numpy as np
import numpy.typing as npt
import structlog
from structlog.stdlib import BoundLogger

from dioptra.sdk.object_detection.bounding_boxes.coordinates import (
    TensorflowBoundingBoxesBatchedGrid,
)
from dioptra.sdk.object_detection.bounding_boxes.postprocessing.bounding_boxes_postprocessing import (  # noqa: B950
    BoundingBoxesYOLOV1PostProcessing,
)

LOGGER: BoundLogger = structlog.stdlib.get_logger()

try:
    import tensorflow as tf
    from tensorflow import Tensor

except ImportError:  # pragma: nocover
    LOGGER.warn(
        "Unable to import one or more optional packages, functionality may be reduced",
        package="tensorflow",
    )


class TensorflowBoundingBoxesYOLOV1Confluence(BoundingBoxesYOLOV1PostProcessing):
    def __init__(
        self,
        bounding_boxes_batched_grid: TensorflowBoundingBoxesBatchedGrid,
        confluence_threshold: float = 0.8,
        score_threshold: float = 0.5,
        min_detection_score: float = 0.5,
        pre_algorithm_threshold: float = 0.05,
        gaussian: bool = False,
        sigma: float = 0.5,
        force_prediction: bool = False,
    ) -> None:
        """
        Args:
            confluence_threshold: value between 0 and 2, with optimum from 0.5-0.8
            score_threshold: class confidence score
            gaussian: boolean switch to turn gaussian decaying of suboptimal bounding
                box confidence scores (setting to False results in suppression of
                suboptimal boxes)
            sigma: used in gaussian decaying. A smaller value causes harsher decaying.

        Returns:
            A dictionary mapping class identity to final retained boxes (and
            corresponding confidence scores)
        """
        self._bbox_batched_grid = bounding_boxes_batched_grid
        self._confluence_threshold = confluence_threshold
        self._score_threshold = score_threshold
        self._min_detection_score = min_detection_score
        self._pre_algorithm_threshold = pre_algorithm_threshold
        self._gaussian = gaussian
        self._sigma = sigma
        self._force_prediction = force_prediction

    @classmethod
    def on_grid_shape(
        cls,
        grid_shape: tuple[int, int],
        confluence_threshold: float = 0.80,
        score_threshold: float = 0.5,
        min_detection_score: float = 0.5,
        pre_algorithm_threshold: float = 0.05,
        gaussian: bool = False,
        sigma: float = 0.5,
        force_prediction: bool = False,
    ) -> TensorflowBoundingBoxesYOLOV1Confluence:
        return cls(
            bounding_boxes_batched_grid=(
                TensorflowBoundingBoxesBatchedGrid.on_grid_shape(grid_shape=grid_shape)
            ),
            confluence_threshold=confluence_threshold,
            score_threshold=score_threshold,
            min_detection_score=min_detection_score,
            pre_algorithm_threshold=pre_algorithm_threshold,
            gaussian=gaussian,
            sigma=sigma,
            force_prediction=force_prediction,
        )

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def postprocess(
        self, bboxes_cell_xywh: Tensor, bboxes_conf: Tensor, bboxes_labels: Tensor
    ) -> tuple[Tensor, Tensor, Tensor, Tensor]:
        batch_size = tf.cast(tf.shape(bboxes_cell_xywh)[0], tf.int32)
        num_boxes = tf.cast(tf.reduce_prod(tf.shape(bboxes_cell_xywh)[1:4]), tf.int32)
        num_labels = tf.cast(tf.shape(bboxes_labels)[-1], tf.int32)

        boxes = tf.reshape(
            self._from_cell_xywh_to_corner(bboxes_cell_xywh=bboxes_cell_xywh),
            shape=(batch_size, num_boxes, 4),
        )

        all_scores = tf.reshape(
            self._calculate_prediction_scores(
                bboxes_conf=bboxes_conf, bboxes_labels=bboxes_labels
            ),
            shape=(batch_size, num_boxes, num_labels),
        )
        top_scores = tf.math.top_k(all_scores, k=1)
        scores = tf.cast(
            tf.reshape(top_scores.values, shape=(batch_size, num_boxes)), tf.float32
        )
        labels = tf.cast(
            tf.reshape(top_scores.indices, shape=(batch_size, num_boxes)), tf.int32
        )

        final_boxes, final_scores, final_labels, final_detections = tf.numpy_function(
            self.confluence,
            [boxes, scores, labels],
            [tf.float32, tf.float32, tf.int32, tf.int32],
        )

        return (
            tf.gather(final_boxes, indices=[1, 0, 3, 2], axis=-1),
            final_scores,
            final_labels,
            final_detections,
        )

    def confluence(
        self,
        boxes: npt.NDArray,
        scores: npt.NDArray,
        labels: npt.NDArray,
    ):
        """
        Args:
            bounding_boxes: list of bounding boxes (x1,y1,x2,y2)
            classes: list of class identifiers (int value, e.g. 1 = person)
            scores: list of class confidence scores (0.0-1.0)

        Returns:
            A dictionary mapping class identity to final retained boxes (and
            corresponding confidence scores)
        """
        batch_size = boxes.shape[0]
        retained_boxes = []
        retained_scores = []
        retained_labels = []

        for batch_idx in range(batch_size):
            class_mapping: dict[int, list[npt.NDArray]] = self.assign_boxes_to_classes(
                boxes[batch_idx], labels[batch_idx], scores[batch_idx]
            )
            output = {}

            for each_class in class_mapping:
                dets = np.array(class_mapping[each_class])
                retain = []

                while dets.size > 0:
                    confluence_scores: list[npt.NDArray] = []
                    proximities = []

                    while len(confluence_scores) < np.size(dets, 0):
                        current_box = len(confluence_scores)
                        x1 = dets[current_box, 0]
                        y1 = dets[current_box, 1]
                        x2 = dets[current_box, 2]
                        y2 = dets[current_box, 3]
                        confidence_score = dets[current_box, 4]

                        xx1 = dets[np.arange(len(dets)) != current_box, 0]
                        yy1 = dets[np.arange(len(dets)) != current_box, 1]
                        xx2 = dets[np.arange(len(dets)) != current_box, 2]
                        yy2 = dets[np.arange(len(dets)) != current_box, 3]
                        cconf = dets[np.arange(len(dets)) != current_box, 4]

                        min_x: npt.NDArray = np.minimum(x1, xx1)
                        min_y: npt.NDArray = np.minimum(y1, yy1)
                        max_x: npt.NDArray = np.maximum(x2, xx2)
                        max_y: npt.NDArray = np.maximum(y2, yy2)

                        x1, y1, x2, y2 = self.normalise_coordinates(
                            x1,
                            y1,
                            x2,
                            y2,
                            min_x,
                            max_x,
                            min_y,
                            max_y,
                        )
                        xx1, yy1, xx2, yy2 = self.normalise_coordinates(
                            xx1,
                            yy1,
                            xx2,
                            yy2,
                            min_x,
                            max_x,
                            min_y,
                            max_y,
                        )

                        hd_x1, hd_x2, vd_y1, vd_y2 = (
                            abs(x1 - xx1),
                            abs(x2 - xx2),
                            abs(y1 - yy1),
                            abs(y2 - yy2),
                        )
                        proximity = hd_x1 + hd_x2 + vd_y1 + vd_y2
                        all_proximities = np.ones_like(proximity)
                        cconf_scores = np.zeros_like(cconf)

                        all_proximities[proximity <= self._confluence_threshold] = (
                            proximity[proximity <= self._confluence_threshold]
                        )
                        cconf_scores[proximity <= self._confluence_threshold] = cconf[
                            proximity <= self._confluence_threshold
                        ]

                        if cconf_scores.size > 0:
                            confluence_score = np.amax(cconf_scores)

                        else:
                            confluence_score = confidence_score

                        if all_proximities.size > 0:
                            proximity = (
                                sum(all_proximities) / all_proximities.size
                            ) * (1 - confidence_score)

                        else:
                            proximity = sum(all_proximities) * (1 - confidence_score)

                        confluence_scores.append(confluence_score)
                        proximities.append(proximity)

                    conf = np.array(confluence_scores)
                    prox = np.array(proximities)

                    dets_temp = np.concatenate((dets, prox[:, None]), axis=1)
                    dets_temp = np.concatenate((dets_temp, conf[:, None]), axis=1)
                    min_idx = np.argmin(dets_temp[:, 5], axis=0)
                    dets[[0, min_idx], :] = dets[[min_idx, 0], :]
                    dets_temp[[0, min_idx], :] = dets_temp[[min_idx, 0], :]
                    dets[0, 4] = dets_temp[0, 6]
                    retain.append(dets[0, :])

                    x1, y1, x2, y2 = dets[0, 0], dets[0, 1], dets[0, 2], dets[0, 3]
                    min_x = np.minimum(x1, dets[1:, 0])
                    min_y = np.minimum(y1, dets[1:, 1])
                    max_x = np.maximum(x2, dets[1:, 2])
                    max_y = np.maximum(y2, dets[1:, 3])

                    x1, y1, x2, y2 = self.normalise_coordinates(
                        x1,
                        y1,
                        x2,
                        y2,
                        min_x,
                        max_x,
                        min_y,
                        max_y,
                    )
                    xx1, yy1, xx2, yy2 = self.normalise_coordinates(
                        dets[1:, 0],
                        dets[1:, 1],
                        dets[1:, 2],
                        dets[1:, 3],
                        min_x,
                        max_x,
                        min_y,
                        max_y,
                    )
                    md_x1, md_x2, md_y1, md_y2 = (
                        abs(x1 - xx1),
                        abs(x2 - xx2),
                        abs(y1 - yy1),
                        abs(y2 - yy2),
                    )
                    manhattan_distance = md_x1 + md_x2 + md_y1 + md_y2
                    weights = np.ones_like(manhattan_distance)

                    if self._gaussian:
                        gaussian_weights = np.exp(
                            -((1 - manhattan_distance) * (1 - manhattan_distance))
                            / self._sigma
                        )
                        weights[manhattan_distance <= self._confluence_threshold] = (
                            gaussian_weights[
                                manhattan_distance <= self._confluence_threshold
                            ]
                        )

                    else:
                        weights[manhattan_distance <= self._confluence_threshold] = (
                            manhattan_distance[
                                manhattan_distance <= self._confluence_threshold
                            ]
                        )

                    dets[1:, 4] *= weights
                    to_reprocess = np.where(dets[1:, 4] >= self._score_threshold)[0]
                    dets = dets[to_reprocess + 1, :]

                output[each_class] = retain

            batch_boxes, batch_scores, batch_labels = self.from_mapping_to_arrays(
                output
            )

            retained_boxes.append(batch_boxes)
            retained_scores.append(batch_scores)
            retained_labels.append(batch_labels)

        return self.pad_retained_arrays(
            boxes=retained_boxes,
            scores=retained_scores,
            labels=retained_labels,
            batch_size=batch_size,
        )

    def assign_boxes_to_classes(
        self, bounding_boxes: npt.NDArray, classes: npt.NDArray, scores: npt.NDArray
    ) -> dict[int, list[npt.NDArray]]:
        """
        Args:
            bounding_boxes: list of bounding boxes (x1,y1,x2,y2)
            classes: list of class identifiers (int value, e.g. 1 = person)
            scores: list of class confidence scores (0.0-1.0)

        Returns:
            defaultdict(list) containing mapping to bounding boxes and confidence scores
            to class
        """
        boxes_to_classes = defaultdict(list)
        for each_box, each_class, each_score in zip(
            bounding_boxes.tolist(), classes.tolist(), scores.tolist()
        ):
            if each_score >= self._pre_algorithm_threshold:
                boxes_to_classes[each_class].append(
                    np.array(
                        [each_box[0], each_box[1], each_box[2], each_box[3], each_score]
                    )
                )

        return boxes_to_classes

    def normalise_coordinates(
        self,
        x1: npt.NDArray,
        y1: npt.NDArray,
        x2: npt.NDArray,
        y2: npt.NDArray,
        min_x: npt.NDArray,
        max_x: npt.NDArray,
        min_y: npt.NDArray,
        max_y: npt.NDArray,
    ) -> tuple[npt.NDArray, npt.NDArray, npt.NDArray, npt.NDArray]:
        """
        Args:
            x1: bounding box coordinate x1
            y1: bounding box coordinate y1
            x2: bounding box coordinate x2
            y2: bounding box coordinate y2
            min_x: minimum bounding box value along x
            max_x: maximum bounding box value along x
            min_y: minimum bounding box value along y
            max_y: maximum bounding box value along y

        Returns:
            Normalised bounding box coordinates (scaled between 0 and 1)
        """
        x1, y1, x2, y2 = (
            (x1 - min_x) / (max_x - min_x),
            (y1 - min_y) / (max_y - min_y),
            (x2 - min_x) / (max_x - min_x),
            (y2 - min_y) / (max_y - min_y),
        )
        return x1, y1, x2, y2

    def from_mapping_to_arrays(
        self, retained
    ) -> tuple[npt.NDArray, npt.NDArray, npt.NDArray]:
        boxes = []
        scores = []
        labels = []

        for label_id, label_boxes in retained.items():
            num_label_boxes = len(label_boxes)
            boxes.extend([x[0:4] for x in label_boxes])
            scores.extend([x[4] for x in label_boxes])
            labels.extend(num_label_boxes * [label_id])

        boxes_array = np.array(boxes).astype("float32")
        scores_array = np.array(scores).astype("float32")
        labels_array = np.array(labels).astype("int32")

        # If no boxes remain after confluence calculation, then return empty arrays
        if scores_array.shape[0] == 0:
            return (
                np.zeros((1, 4)).astype("float32"),
                np.zeros(1).astype("float32"),
                np.zeros(1).astype("int32"),
            )

        # If one or more boxes remain after confluence calculation, then drop all below
        # min detection score
        scores_threshold_mask = scores_array >= self._min_detection_score
        num_retained = np.sum(scores_threshold_mask)

        if num_retained > 0:
            return (
                boxes_array[scores_threshold_mask],
                scores_array[scores_threshold_mask],
                labels_array[scores_threshold_mask],
            )

        # If the min detection score threshold drops all the remaining boxes, then check
        # if force prediction option is enabled If force prediction is disabled, then
        # just return empty arrays
        if not self._force_prediction:
            return (
                np.zeros((1, 4)).astype("float32"),
                np.zeros(1).astype("float32"),
                np.zeros(1).astype("int32"),
            )

        # If force prediction is enabled, then keep the best prediction and drop the
        # rest
        idx = int(np.argmax(scores_array))
        return (
            boxes_array[idx : idx + 1],
            scores_array[idx : idx + 1],
            labels_array[idx : idx + 1],
        )

    def pad_retained_arrays(
        self, boxes, scores, labels, batch_size
    ) -> tuple[npt.NDArray, npt.NDArray, npt.NDArray, npt.NDArray]:
        padded_boxes: npt.NDArray = np.zeros(
            (batch_size, max(max([x.shape[0] for x in boxes]), 1), 4), dtype="float32"
        )
        padded_scores: npt.NDArray = np.zeros(
            (batch_size, max(max([x.shape[0] for x in scores]), 1)), dtype="float32"
        )
        padded_labels: npt.NDArray = np.zeros(
            (batch_size, max(max([x.shape[0] for x in labels]), 1)), dtype="int32"
        )
        padded_detections: npt.NDArray = np.zeros(batch_size, dtype="int32")

        for batch_idx, (batch_boxes, batch_scores, batch_labels) in enumerate(
            zip(boxes, scores, labels)
        ):
            if len(batch_boxes) > 0:
                padded_boxes[batch_idx, 0 : batch_boxes.shape[0]] = batch_boxes.astype(
                    "float32"
                )

            if len(batch_scores) > 0:
                padded_scores[batch_idx, 0 : batch_scores.shape[0]] = (
                    batch_scores.astype("float32")
                )

            if len(batch_labels) > 0:
                padded_labels[batch_idx, 0 : batch_labels.shape[0]] = (
                    batch_labels.astype("int32")
                )

            padded_detections[batch_idx] = len(batch_scores)

        return padded_boxes, padded_scores, padded_labels, padded_detections

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.int32),
            tf.TensorSpec(None, tf.int32),
        ]
    )
    def embed(
        self, bboxes_corner: Tensor, bboxes_labels: Tensor, n_classes: Tensor
    ) -> tuple[Tensor, Tensor, Tensor, Tensor]:
        return cast(
            tuple[Tensor, Tensor, Tensor, Tensor],
            self._bbox_batched_grid.embed(
                bboxes_corner=bboxes_corner,
                bboxes_labels=bboxes_labels,
                n_classes=n_classes,
            ),
        )

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def _from_cell_xywh_to_corner(self, bboxes_cell_xywh: Tensor) -> Tensor:
        n_bounding_boxes = tf.cast(tf.shape(bboxes_cell_xywh)[-2], tf.int32)

        return self._bbox_batched_grid.from_cell_xywh_to_corner(
            bboxes_cell_xywh=bboxes_cell_xywh,
            n_bounding_boxes=n_bounding_boxes,
        )

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def _calculate_prediction_scores(
        self, bboxes_conf: Tensor, bboxes_labels: Tensor
    ) -> Tensor:
        n_bounding_boxes = tf.cast(tf.shape(bboxes_conf)[-1], tf.int32)
        bboxes_labels = tf.tile(
            tf.expand_dims(bboxes_labels, axis=-2),
            multiples=(1, 1, 1, n_bounding_boxes, 1),
        )

        return tf.expand_dims(bboxes_conf, axis=-1) * bboxes_labels



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

from typing import Optional, cast

import structlog
from structlog.stdlib import BoundLogger

LOGGER: BoundLogger = structlog.stdlib.get_logger()


def mock_s3_upload(self, *args, **kwargs) -> str | None:
    LOGGER.info(
        "Mocking S3Service.upload() function",
        args=args,
        kwargs=kwargs,
    )
    fileobj = kwargs.get("fileobj")
    bucket = kwargs.get("bucket")
    key = kwargs.get("key")

    log: BoundLogger = kwargs.get("log", LOGGER.new())
    log.info("Uploading data to S3", fileobj=fileobj, bucket=bucket, key=key)
    uri = cast(Optional[str], self.as_uri(bucket=bucket, key=key))
    log.info("S3 upload successful", uri=uri)

    return uri



from dioptra.sdk.utilities.paths import clear_directory


def test_clear_directory(tmp_path):
    files = ["file1", "dir1/file2", "dir1/dir2/file3"]

    # Make some files
    for f in files:
        # Ensure we're not trying to create files in the root directory!
        f = f.lstrip("/")
        pathobj = tmp_path / f
        pathobj.parent.mkdir(parents=True, exist_ok=True)
        pathobj.touch()

    # Test the function
    clear_directory(tmp_path)

    # Ensure the directory is now clear!
    test_files = list(tmp_path.iterdir())
    assert len(test_files) == 0



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""A module for binding configurations to shared services using dependency injection."""
from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Any, Callable, List, Optional

from boto3.session import Session
from botocore.client import BaseClient
from injector import Binder, Module, provider
from mlflow.tracking import MlflowClient
from passlib.context import CryptContext
from redis import Redis

from dioptra.restapi.v1.shared.password_service import PasswordService
from dioptra.restapi.v1.shared.request_scope import request
from dioptra.restapi.v1.shared.rq_service import RQServiceV1


class MLFlowClientModule(Module):
    @request
    @provider
    def provide_mlflow_client_module(
        self,
    ) -> MlflowClient:
        return MlflowClient()


@dataclass
class RQServiceConfiguration(object):
    redis: Redis


class RQServiceV1Module(Module):
    @request
    @provider
    def provide_rq_service_module(
        self, configuration: RQServiceConfiguration
    ) -> RQServiceV1:
        return RQServiceV1(redis=configuration.redis)


@dataclass
class PasswordServiceConfiguration(object):
    crypt_context: CryptContext


class PasswordServiceModule(Module):
    @provider
    def provide_password_service_module(
        self, configuration: PasswordServiceConfiguration
    ) -> PasswordService:
        return PasswordService(crypt_context=configuration.crypt_context)


def _bind_rq_service_configuration(binder: Binder):
    redis_conn: Redis = Redis.from_url(os.getenv("RQ_REDIS_URI", "redis://"))
    configuration: RQServiceConfiguration = RQServiceConfiguration(redis=redis_conn)
    binder.bind(RQServiceConfiguration, to=configuration, scope=request)


def _bind_s3_service_configuration(binder: Binder) -> None:
    s3_endpoint_url: Optional[str] = os.getenv("MLFLOW_S3_ENDPOINT_URL")

    s3_session: Session = Session()
    s3_client: BaseClient = s3_session.client("s3", endpoint_url=s3_endpoint_url)

    binder.bind(Session, to=s3_session, scope=request)
    binder.bind(BaseClient, to=s3_client, scope=request)


def _bind_password_service_configuration(binder: Binder):
    configuration: PasswordServiceConfiguration = PasswordServiceConfiguration(
        crypt_context=CryptContext(
            schemes=["pbkdf2_sha256"],
            pbkdf2_sha256__default_rounds=30000,
        ),
    )

    binder.bind(PasswordServiceConfiguration, to=configuration)


def bind_dependencies(binder: Binder) -> None:
    """Binds interfaces to implementations within the main application.

    Args:
        binder: A :py:class:`~injector.Binder` object.
    """

    _bind_rq_service_configuration(binder)
    _bind_s3_service_configuration(binder)
    _bind_password_service_configuration(binder)


def register_providers(modules: List[Callable[..., Any]]) -> None:
    """Registers type providers within the main application.

    Args:
        modules: A list of callables used for configuring the dependency injection
            environment.
    """

    modules.append(MLFlowClientModule)
    modules.append(RQServiceV1Module)
    modules.append(PasswordServiceModule)



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The module defining the endpoints for Drafts."""
import uuid
from functools import partial
from typing import Type, cast

import structlog
from flask import request
from flask_accepts import accepts, responds
from flask_login import login_required
from flask_restx import Namespace, Resource
from injector import ClassAssistedBuilder, inject
from marshmallow import Schema
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import models
from dioptra.restapi.v1 import utils
from dioptra.restapi.v1.schemas import IdStatusResponseSchema

from .schema import DraftGetQueryParameters, DraftPageSchema, DraftSchema
from .service import (
    ResourceDraftsIdService,
    ResourceDraftsService,
    ResourceIdDraftService,
)

LOGGER: BoundLogger = structlog.stdlib.get_logger()


def generate_resource_drafts_endpoint(
    api: Namespace,
    resource_name: str,
    route_prefix: str,
    request_schema: Type[Schema],
) -> Resource:
    """
    Generates an Resource class for creating and retrieving Drafts

    Args:
        api: The API namespace
        resource_name: The name of the resource
        request_schema: The request schema type

    Returns:
        The generated Resource class
    """

    # Based on: https://github.com/apryor6/flask_accepts/blob/05567461c421a534d6fc6e122d5e086b0b0e53aa/flask_accepts/utils.py#L154-L160  # noqa: B950
    if isinstance(request_schema, Schema):
        model_name = "Drafts" + "".join(
            request_schema.__class__.__name__.rsplit("Schema", 1)
        )
    else:
        model_name = "Drafts" + "".join(request_schema.__name__.rsplit("Schema", 1))

    @api.route("/drafts/")
    class ResourceDraftsEndpoint(Resource):
        @inject
        def __init__(
            self,
            draft_service: ClassAssistedBuilder[ResourceDraftsService],
            *args,
            **kwargs,
        ) -> None:
            self._draft_service = draft_service.build(resource_type=resource_name)
            super().__init__(*args, **kwargs)

        @login_required
        @accepts(query_params_schema=DraftGetQueryParameters, api=api)
        @responds(schema=DraftPageSchema, api=api)
        def get(self):
            """Gets the Drafts for the resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Draft", request_type="GET"
            )
            parsed_query_params = request.parsed_query_params  # noqa: F841

            draft_type = parsed_query_params["draft_type"].value
            group_id = parsed_query_params["group_id"]
            page_index = parsed_query_params["index"]
            page_length = parsed_query_params["page_length"]

            drafts, total_num_drafts = self._draft_service.get(
                draft_type=draft_type,
                group_id=group_id,
                base_resource_id=None,
                page_index=page_index,
                page_length=page_length,
                log=log,
            )
            return utils.build_paging_envelope(
                f"{route_prefix}/drafts",
                build_fn=partial(
                    utils.build_resource_draft, draft_schema=request_schema
                ),
                data=drafts,
                group_id=group_id,
                query=None,
                draft_type=draft_type,
                index=page_index,
                length=page_length,
                total_num_elements=total_num_drafts,
            )

        @login_required
        @accepts(schema=request_schema, model_name=model_name, api=api)
        @responds(schema=DraftSchema, api=api)
        def post(self):
            """Creates a Draft for the resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Draft", request_type="POST"
            )
            parsed_obj = request.parsed_obj  # noqa: F841
            draft = self._draft_service.create(None, parsed_obj, log=log)
            return utils.build_resource_draft(draft, draft_schema=request_schema)

    return ResourceDraftsEndpoint


def generate_resource_drafts_id_endpoint(
    api: Namespace, resource_name: str, request_schema: Type[Schema]
) -> Resource:
    """
    Generates an Resource class for modifying and deleting Drafts

    Args:
        api: The API namespace
        resource_name: The name of the resource
        request_schema: The request schema type

    Returns:
        The generated Resource class
    """

    # Based on: https://github.com/apryor6/flask_accepts/blob/05567461c421a534d6fc6e122d5e086b0b0e53aa/flask_accepts/utils.py#L154-L160  # noqa: B950
    if isinstance(request_schema, Schema):
        model_name = "DraftsId" + "".join(
            request_schema.__class__.__name__.rsplit("Schema", 1)
        )
    else:
        model_name = "DraftsId" + "".join(request_schema.__name__.rsplit("Schema", 1))

    @api.route("/drafts/<int:draftId>")
    @api.param("draftId", f"ID for the Draft of the {resource_name} resource.")
    class ResourcesDraftsIdEndpoint(Resource):
        @inject
        def __init__(
            self,
            draft_id_service: ClassAssistedBuilder[ResourceDraftsIdService],
            *args,
            **kwargs,
        ) -> None:
            self._draft_id_service = draft_id_service.build(resource_type=resource_name)
            super().__init__(*args, **kwargs)

        @login_required
        @responds(schema=DraftSchema, api=api)
        def get(self, draftId: int):
            """Gets a Draft for the resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Draft", request_type="GET"
            )
            draft = self._draft_id_service.get(
                draftId, error_if_not_found=True, log=log
            )
            return utils.build_resource_draft(
                cast(models.DraftResource, draft), request_schema
            )

        @login_required
        @accepts(schema=request_schema, model_name=model_name, api=api)
        @responds(schema=DraftSchema, api=api)
        def put(self, draftId: int):
            """Modifies a Draft for the resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Draft", request_type="POST"
            )
            parsed_obj = request.parsed_obj  # type: ignore
            draft = self._draft_id_service.modify(
                draftId, payload=parsed_obj, log=log, error_if_not_found=True
            )
            return utils.build_resource_draft(
                cast(models.DraftResource, draft), request_schema
            )

        @login_required
        @responds(schema=IdStatusResponseSchema, api=api)
        def delete(self, draftId: int):
            """Deletes a Draft for the resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Draft", request_type="DELETE"
            )
            return self._draft_id_service.delete(draftId, log=log)

    return ResourcesDraftsIdEndpoint


def generate_resource_id_draft_endpoint(
    api: Namespace, resource_name: str, request_schema: Type[Schema]
) -> Resource:
    """
    Generates an Resource class for managing the Draft of an existing resource.

    Args:
        api: The API namespace
        resource_name: The name of the resource
        request_schema: The request schema type

    Returns:
        The generated Resource class
    """

    # Based on: https://github.com/apryor6/flask_accepts/blob/05567461c421a534d6fc6e122d5e086b0b0e53aa/flask_accepts/utils.py#L154-L160  # noqa: B950
    if isinstance(request_schema, Schema):
        model_name = "Draft" + "".join(
            request_schema.__class__.__name__.rsplit("Schema", 1)
        )
    else:
        model_name = "Draft" + "".join(request_schema.__name__.rsplit("Schema", 1))

    @api.route("/<int:id>/draft")
    @api.param("id", "ID for the resource.")
    class ResourcesIdDraftEndpoint(Resource):
        @inject
        def __init__(
            self,
            id_draft_service: ClassAssistedBuilder[ResourceIdDraftService],
            *args,
            **kwargs,
        ) -> None:
            self._id_draft_service = id_draft_service.build(resource_type=resource_name)
            super().__init__(*args, **kwargs)

        @login_required
        @responds(schema=DraftSchema, api=api)
        def get(self, id: int):
            """Gets the Draft for this resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Draft", request_type="GET"
            )
            draft, num_other_drafts = self._id_draft_service.get(
                id, error_if_not_found=True, log=log
            )
            return utils.build_resource_draft(
                cast(models.DraftResource, draft), request_schema, num_other_drafts
            )

        @login_required
        @accepts(schema=request_schema, model_name=model_name, api=api)
        @responds(schema=DraftSchema, api=api)
        def post(self, id: int):
            """Creates a Draft for this resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Draft", request_type="POST"
            )
            parsed_obj = request.parsed_obj  # type: ignore
            draft, num_other_drafts = self._id_draft_service.create(
                id, base_resource_id=None, payload=parsed_obj, log=log
            )
            return utils.build_resource_draft(
                cast(models.DraftResource, draft), request_schema, num_other_drafts
            )

        @login_required
        @accepts(schema=request_schema, model_name=model_name, api=api)
        @responds(schema=DraftSchema, api=api)
        def put(self, id: int):
            """Modifies the Draft for this resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Draft", request_type="POST"
            )
            parsed_obj = request.parsed_obj  # type: ignore
            draft, num_other_drafts = self._id_draft_service.modify(
                id, payload=parsed_obj, error_if_not_found=True, log=log
            )
            return utils.build_resource_draft(
                cast(models.DraftResource, draft), request_schema, num_other_drafts
            )

        @login_required
        @responds(schema=IdStatusResponseSchema, api=api)
        def delete(self, id: int):
            """Deletes the Draft for this resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Draft", request_type="DELETE"
            )
            return self._id_draft_service.delete(id, log=log)

    return ResourcesIdDraftEndpoint


def generate_nested_resource_drafts_endpoint(
    api: Namespace,
    resource_name: str,
    resource_route: str,
    base_resource_route: str,
    request_schema: Type[Schema],
) -> Resource:
    """
    Generates an Resource class for creating and retrieving Drafts

    Args:
        api: The API namespace
        resource_name: The name of the nested resource
        resource_route: The route of the nested resource
        base_resource_route: The route of the base resource
        request_schema: The request schema type

    Returns:
        The generated Resource class
    """

    # Based on: https://github.com/apryor6/flask_accepts/blob/05567461c421a534d6fc6e122d5e086b0b0e53aa/flask_accepts/utils.py#L154-L160  # noqa: B950
    if isinstance(request_schema, Schema):
        model_name = "NestedDrafts" + "".join(
            request_schema.__class__.__name__.rsplit("Schema", 1)
        )
    else:
        model_name = "NestedDrafts" + "".join(
            request_schema.__name__.rsplit("Schema", 1)
        )

    @api.route(f"/<int:id>/{resource_route}/drafts/")
    class ResourceDraftsEndpoint(Resource):
        @inject
        def __init__(
            self,
            draft_service: ClassAssistedBuilder[ResourceDraftsService],
            *args,
            **kwargs,
        ) -> None:
            self._draft_service = draft_service.build(resource_type=resource_name)
            super().__init__(*args, **kwargs)

        @login_required
        @accepts(query_params_schema=DraftGetQueryParameters, api=api)
        @responds(schema=DraftPageSchema, api=api)
        def get(self, id: int):
            """Gets the Drafts for the resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Draft", request_type="GET"
            )
            parsed_query_params = request.parsed_query_params  # type: ignore

            draft_type = parsed_query_params["draft_type"].value
            group_id = parsed_query_params["group_id"]
            page_index = parsed_query_params["index"]
            page_length = parsed_query_params["page_length"]

            drafts, total_num_drafts = self._draft_service.get(
                draft_type=draft_type,
                group_id=group_id,
                base_resource_id=id,
                page_index=page_index,
                page_length=page_length,
                log=log,
            )
            return utils.build_paging_envelope(
                f"{base_resource_route}/{id}/{resource_route}/drafts",
                build_fn=partial(
                    utils.build_resource_draft, draft_schema=request_schema
                ),
                data=drafts,
                group_id=group_id,
                query=None,
                draft_type=draft_type,
                index=page_index,
                length=page_length,
                total_num_elements=total_num_drafts,
            )

        @login_required
        @accepts(schema=request_schema, model_name=model_name, api=api)
        @responds(schema=DraftSchema, api=api)
        def post(self, id: int):
            """Creates a Draft for the resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Draft", request_type="POST"
            )
            parsed_obj = request.parsed_obj  # type: ignore
            draft = self._draft_service.create(id, parsed_obj, log=log)
            return utils.build_resource_draft(draft, draft_schema=request_schema)

    return ResourceDraftsEndpoint


def generate_nested_resource_drafts_id_endpoint(
    api: Namespace,
    resource_name: str,
    resource_route: str,
    request_schema: Type[Schema],
) -> Resource:
    """
    Generates an Resource class for modifying and deleting Drafts

    Args:
        api: The API namespace
        resource_name: The name of the resource
        resource_route: The route of the resource
        request_schema: The request schema type

    Returns:
        The generated Resource class
    """

    # Based on: https://github.com/apryor6/flask_accepts/blob/05567461c421a534d6fc6e122d5e086b0b0e53aa/flask_accepts/utils.py#L154-L160  # noqa: B950
    if isinstance(request_schema, Schema):
        model_name = "NestedDraftsId" + "".join(
            request_schema.__class__.__name__.rsplit("Schema", 1)
        )
    else:
        model_name = "NestedDraftsId" + "".join(
            request_schema.__name__.rsplit("Schema", 1)
        )

    @api.route(f"/<int:id>/{resource_route}/drafts/<int:draftId>")
    @api.param("draftId", f"ID for the Draft of the {resource_name} resource.")
    class ResourcesDraftsIdEndpoint(Resource):
        @inject
        def __init__(
            self,
            draft_id_service: ClassAssistedBuilder[ResourceDraftsIdService],
            *args,
            **kwargs,
        ) -> None:
            self._draft_id_service = draft_id_service.build(resource_type=resource_name)
            super().__init__(*args, **kwargs)

        @login_required
        @responds(schema=DraftSchema, api=api)
        def get(self, id: int, draftId: int):
            """Gets a Draft for the resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Draft", request_type="GET"
            )
            draft = self._draft_id_service.get(
                draftId, error_if_not_found=True, log=log
            )
            return utils.build_resource_draft(
                cast(models.DraftResource, draft), draft_schema=request_schema
            )

        @login_required
        @accepts(schema=request_schema, model_name=model_name, api=api)
        @responds(schema=DraftSchema, api=api)
        def put(self, id: int, draftId: int):
            """Modifies a Draft for the resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Draft", request_type="POST"
            )
            parsed_obj = request.parsed_obj  # type: ignore
            draft = self._draft_id_service.modify(draftId, payload=parsed_obj, log=log)
            return utils.build_resource_draft(
                cast(models.DraftResource, draft), draft_schema=request_schema
            )

        @login_required
        @responds(schema=IdStatusResponseSchema, api=api)
        def delete(self, id: int, draftId: int):
            """Deletes a Draft for the resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Draft", request_type="DELETE"
            )
            return self._draft_id_service.delete(draftId, log=log)

    return ResourcesDraftsIdEndpoint


def generate_nested_resource_id_draft_endpoint(
    api: Namespace,
    resource_name: str,
    resource_route: str,
    request_schema: Type[Schema],
) -> Resource:
    """
    Generates an Resource class for managing the Draft of an existing resource.

    Args:
        api: The API namespace
        resource_name: The name of the resource
        resource_route: The route of the resource
        request_schema: The request schema type

    Returns:
        The generated Resource class
    """

    route_singular = resource_route[:-1]
    resource_id = f"{route_singular}Id"

    # Based on: https://github.com/apryor6/flask_accepts/blob/05567461c421a534d6fc6e122d5e086b0b0e53aa/flask_accepts/utils.py#L154-L160  # noqa: B950
    if isinstance(request_schema, Schema):
        model_name = "NestedDraft" + "".join(
            request_schema.__class__.__name__.rsplit("Schema", 1)
        )
    else:
        model_name = "NestedDraft" + "".join(
            request_schema.__name__.rsplit("Schema", 1)
        )

    @api.route(f"/<int:id>/{resource_route}/<int:{resource_id}>/draft")
    @api.param("id", "ID for the resource.")
    class ResourcesIdDraftEndpoint(Resource):
        @inject
        def __init__(
            self,
            id_draft_service: ClassAssistedBuilder[ResourceIdDraftService],
            *args,
            **kwargs,
        ) -> None:
            self._id_draft_service = id_draft_service.build(resource_type=resource_name)
            super().__init__(*args, **kwargs)

        @login_required
        @responds(schema=DraftSchema, api=api)
        def get(self, id: int, **kwargs):
            """Gets the Draft for this resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Draft", request_type="GET"
            )
            if set(kwargs.keys()) != set([resource_id]):
                unexpected_kwargs = {
                    k: v for k, v in kwargs.items() if resource_id != k
                }
                log.error("Unexpected input", kwargs=unexpected_kwargs)
                raise TypeError(
                    "delete method received unexpected keyword arguments: "
                    f"{list(unexpected_kwargs.keys())}"
                )

            draft, num_other_drafts = self._id_draft_service.get(
                kwargs[resource_id], error_if_not_found=True, log=log
            )
            return utils.build_resource_draft(
                cast(models.DraftResource, draft), request_schema, num_other_drafts
            )

        @login_required
        @accepts(schema=request_schema, model_name=model_name, api=api)
        @responds(schema=DraftSchema, api=api)
        def post(self, id: int, **kwargs):
            """Creates a Draft for this resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Draft", request_type="POST"
            )
            if set(kwargs.keys()) != set([resource_id]):
                unexpected_kwargs = {
                    k: v for k, v in kwargs.items() if resource_id != k
                }
                log.error("Unexpected input", kwargs=unexpected_kwargs)
                raise TypeError(
                    "delete method received unexpected keyword arguments: "
                    f"{list(unexpected_kwargs.keys())}"
                )

            parsed_obj = request.parsed_obj  # type: ignore
            draft, num_other_drafts = self._id_draft_service.create(
                kwargs[resource_id], base_resource_id=id, payload=parsed_obj, log=log
            )
            return utils.build_resource_draft(
                cast(models.DraftResource, draft), request_schema, num_other_drafts
            )

        @login_required
        @accepts(schema=request_schema, model_name=model_name, api=api)
        @responds(schema=DraftSchema, api=api)
        def put(self, id: int, **kwargs):
            """Modifies the Draft for this resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Draft", request_type="POST"
            )
            if set(kwargs.keys()) != set([resource_id]):
                unexpected_kwargs = {
                    k: v for k, v in kwargs.items() if resource_id != k
                }
                log.error("Unexpected input", kwargs=unexpected_kwargs)
                raise TypeError(
                    "delete method received unexpected keyword arguments: "
                    f"{list(unexpected_kwargs.keys())}"
                )

            parsed_obj = request.parsed_obj  # type: ignore
            draft, num_other_drafts = self._id_draft_service.modify(
                kwargs[resource_id],
                payload=parsed_obj,
                error_if_not_found=True,
                log=log,
            )
            return utils.build_resource_draft(
                cast(models.DraftResource, draft), request_schema, num_other_drafts
            )

        @login_required
        @responds(schema=IdStatusResponseSchema, api=api)
        def delete(self, id: int, **kwargs):
            """Deletes the Draft for this resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Draft", request_type="DELETE"
            )
            if set(kwargs.keys()) != set([resource_id]):
                unexpected_kwargs = {
                    k: v for k, v in kwargs.items() if resource_id != k
                }
                log.error("Unexpected input", kwargs=unexpected_kwargs)
                raise TypeError(
                    "delete method received unexpected keyword arguments: "
                    f"{list(unexpected_kwargs.keys())}"
                )

            return self._id_draft_service.delete(kwargs[resource_id], log=log)

    return ResourcesIdDraftEndpoint



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import logging
import sys
from logging import getLogger
from typing import Any, Callable, List, Mapping, MutableMapping, Optional, Tuple, Union

import structlog

ProcessorType = Callable[
    [Any, str, MutableMapping[str, Any]],
    Union[Mapping[str, Any], str, bytes, Tuple[Any, ...]],
]


def attach_stdout_stream_handler(
    as_json: bool, logger: Optional[logging.Logger] = None
) -> None:
    logger = logger or getLogger()
    log_processor: ProcessorType = _get_structlog_processor(as_json)

    handler = logging.StreamHandler(sys.stdout)
    formatter = structlog.stdlib.ProcessorFormatter(
        processor=log_processor,
        foreign_pre_chain=[
            structlog.stdlib.add_logger_name,
            structlog.stdlib.add_log_level,
            structlog.processors.TimeStamper(fmt="iso", utc=True),
            structlog.processors.StackInfoRenderer(),
            structlog.processors.format_exc_info,
            structlog.processors.UnicodeDecoder(),
        ],
    )
    handler.setFormatter(formatter)
    logging.captureWarnings(True)
    logger.addHandler(handler)


def clear_logger_handlers(logger: Optional[logging.Logger]) -> None:
    if logger is None:
        return None

    for handler in logger.handlers:
        logger.removeHandler(handler)


def configure_structlog() -> None:
    processors: List[Any] = [
        structlog.stdlib.filter_by_level,
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        structlog.stdlib.PositionalArgumentsFormatter(),
        structlog.processors.TimeStamper(fmt="iso", utc=True),
        structlog.processors.StackInfoRenderer(),
        structlog.processors.format_exc_info,
        structlog.processors.UnicodeDecoder(),
        structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
    ]

    structlog.configure_once(
        processors=processors,
        context_class=dict,
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )


def set_logging_level(level: str, logger: Optional[logging.Logger] = None) -> None:
    logger = logger or getLogger()
    logger.setLevel(_get_logging_level(level.strip().upper()))


def _get_structlog_processor(as_json: bool) -> ProcessorType:
    if as_json:
        log_processor: ProcessorType = structlog.processors.JSONRenderer()
        return log_processor

    log_processor = structlog.dev.ConsoleRenderer()
    return log_processor


def _get_logging_level(level: str) -> str:
    allowed_levels = {"DEBUG", "ERROR", "INFO", "WARNING"}

    if level not in allowed_levels:
        level = "INFO"

    return level



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
#
# The LogStream class is adapted from the following source:
#
#     J. Paton, "Redirect standard out to Python’s logging module with contextlib,"
#         John Paton, May 22, 2019. https://johnpaton.net/posts/redirect-logging/
#         (accessed Jan. 18, 2021).

import contextlib
import logging
from abc import ABCMeta, abstractmethod


class LogStream(metaclass=ABCMeta):
    @abstractmethod
    def __init__(self, as_json: bool) -> None:
        pass

    def write(self, msg):
        if msg and not msg.isspace():
            self.logger.log(self.level, self._format_newlines(msg))

    @abstractmethod
    def close(self):
        pass

    @abstractmethod
    def flush(self):
        pass

    def _format_newlines(self, msg):
        if self._as_json:
            return "\n".join([x.rstrip() for x in msg.rstrip().splitlines()])

        return "||".join([x.rstrip() for x in msg.rstrip().splitlines()])

    def __enter__(self):
        self._redirector.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._redirector.__exit__(exc_type, exc_value, traceback)


class StdoutLogStream(LogStream):
    def __init__(self, as_json: bool):
        self.logger = logging.getLogger("STDOUT")
        self.name = self.logger.name
        self.level = logging.INFO
        self._as_json = as_json
        self._redirector = contextlib.redirect_stdout(self)  # type: ignore

    def close(self):
        pass

    def flush(self):
        pass


class StderrLogStream(LogStream):
    def __init__(self, as_json: bool):
        self.logger = logging.getLogger("STDERR")
        self.name = self.logger.name
        self.level = logging.INFO
        self._as_json = as_json
        self._redirector = contextlib.redirect_stderr(self)  # type: ignore

    def close(self):
        pass

    def flush(self):
        pass



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

import numpy as np
import pytest


@pytest.mark.parametrize(
    "request_",
    [
        [],
        [
            {"name": "a", "func": "accuracy"},
            {"name": "b", "func": "precision"},
            {"name": "c", "func": "roc_auc"},
            {"name": "d", "func": "categorical_accuracy"},
            {"name": "e", "func": "mcc"},
            {"name": "f", "func": "f1"},
            {"name": "g", "func": "recall"},
        ],
        [{"name": "a", "func": "recall"}, {"name": "b", "func": "f1"}],
        [{"name": "c", "func": "precision"}, {"name": "d", "func": "recall"}],
        [{"name": "e", "func": "accuracy"}],
    ],
)
@pytest.mark.parametrize(
    ("y_true", "y_pred"),
    [
        ([0, 1], [1, 1]),
        ([1, 0, 0], [1, 1, 1]),
        ([1, 0, 0, 0, 1, 1], [0, 1, 1, 1, 0, 0]),
        ([1, 0, 1], [1, 0, 0]),
        ([0, 1, 1, 0, 1, 0, 1], [1, 0, 0, 1, 0, 0, 0]),
    ],
)
def test_get_performance_metric_list(request_, y_true, y_pred) -> None:
    from dioptra_builtins.metrics.performance import get_performance_metric_list

    y_true = np.array(y_true)
    y_pred = np.array(y_pred)

    result: list = get_performance_metric_list(request_)
    for res in result:
        assert isinstance(res[0], str)
        performance = res[1](y_true, y_pred)
        assert isinstance(performance, float)


@pytest.mark.parametrize(
    "func",
    ["accuracy", "precision", "roc_auc", "categorical_accuracy", "mcc", "f1", "recall"],
)
@pytest.mark.parametrize(
    ("y_true", "y_pred"),
    [
        ([0, 1], [1, 1]),
        ([1, 0, 0], [1, 1, 1]),
        ([1, 0, 0, 0, 1, 1], [0, 1, 1, 1, 0, 0]),
        ([1, 0, 1], [1, 0, 0]),
        ([0, 1, 1, 0, 1, 0, 1], [1, 0, 0, 1, 0, 0, 0]),
    ],
)
def test_get_performance_metric(func, y_true, y_pred) -> None:
    from dioptra_builtins.metrics.performance import get_performance_metric

    y_true = np.array(y_true)
    y_pred = np.array(y_pred)

    metric = get_performance_metric(func)
    result = metric(y_true, y_pred)
    assert isinstance(result, float)


@pytest.mark.parametrize(
    ("y_true", "y_pred", "weights", "normalize", "expected"),
    [
        ([0, 1], [1, 1], None, True, 0.5),
        ([1, 2, 4], [4, 1, 1], [0.3, 0.6, 0.1], False, 0.0),
        ([1, 2, 3, 6, 5, 6], [6, 6, 6, 6, 6, 6], None, True, 0.33333333),
        (
            [9, 9, 9, 9, 9, 2],
            [9, 9, 9, 9, 9, 9],
            [0.2, 0.1, 0.3, 0.1, 0.1, 0.2],
            False,
            0.8,
        ),
        ([2, 2, 3, 4, 5], [2, 2, 5, 5, 5], [0.2, 0.2, 0.2, 0.2, 0.2], True, 0.6),
    ],
)
def test_accuracy(y_true, y_pred, weights, normalize, expected) -> None:
    from dioptra_builtins.metrics.performance import accuracy

    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    expected = np.array(expected)
    kwargs = {"normalize": normalize, "sample_weight": weights}
    result: float = accuracy(y_true, y_pred, **kwargs)
    assert np.isclose(result, expected)


@pytest.mark.parametrize(
    ("y_true", "y_pred", "weights", "average", "multiclass", "expected"),
    [
        (
            [2, 1, 3],
            [[0.3, 0.5, 0.2], [0.4, 0.5, 0.1], [0.3, 0.3, 0.4]],
            None,
            "macro",
            "ovr",
            0.9166666,
        ),
        (
            [2, 1, 3, 1, 1],
            [
                [0.3, 0.5, 0.2],
                [0.4, 0.5, 0.1],
                [0.3, 0.3, 0.4],
                [0.9, 0.05, 0.05],
                [0.1, 0.1, 0.8],
            ],
            None,
            "macro",
            "ovo",
            0.8055555555555557,
        ),
        (
            [2, 1, 3, 1, 1],
            [
                [0.3, 0.5, 0.2],
                [0.4, 0.5, 0.1],
                [0.3, 0.3, 0.4],
                [0.9, 0.05, 0.05],
                [0.1, 0.1, 0.8],
            ],
            [0.2, 0.3, 0.3, 0.1, 0.1],
            "macro",
            "ovr",
            0.8232142857142858,
        ),
        (
            [2, 1, 3, 1, 1],
            [
                [0.3, 0.5, 0.2],
                [0.4, 0.5, 0.1],
                [0.3, 0.3, 0.4],
                [0.9, 0.05, 0.05],
                [0.1, 0.1, 0.8],
            ],
            None,
            "weighted",
            "ovo",
            0.7666666666666668,
        ),
        (
            [2, 1, 3, 1, 1],
            [
                [0.1, 0.5, 0.4],
                [0.4, 0.5, 0.1],
                [0, 0.9, 0.1],
                [0.9, 0.05, 0.05],
                [0.1, 0.1, 0.8],
            ],
            None,
            "weighted",
            "ovr",
            0.75,
        ),
        ([0, 1, 0, 1, 1], [1, 0, 1, 0, 1], None, "micro", "ovo", 0.1666666),
        ([0, 1, 0, 1, 0], [1, 0, 1, 1, 1], None, "macro", "ovr", 0.25),
        ([1, 1, 1, 1, 0], [1, 1, 1, 1, 0], None, "weighted", "ovr", 1.0),
    ],
)
def test_roc_auc(y_true, y_pred, weights, average, multiclass, expected) -> None:
    from dioptra_builtins.metrics.performance import roc_auc

    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    kwargs = {"average": average, "multi_class": multiclass, "sample_weight": weights}
    result: float = roc_auc(y_true, y_pred, **kwargs)
    assert np.isclose(result, expected)


@pytest.mark.parametrize(
    ("y_true", "y_pred", "expected"),
    [
        ([0, 1], [1, 1], 0.5),
        ([1, 2, 4], [4, 1, 1], 0.0),
        ([1, 2, 3, 6, 5, 6], [6, 6, 6, 6, 6, 6], 0.33333333),
        ([9, 9, 9, 9, 9, 2], [9, 9, 9, 9, 9, 9], 0.83333333),
        ([2, 2, 3, 4, 5], [2, 2, 5, 5, 5], 0.6),
    ],
)
def test_categorical_accuracy(y_true, y_pred, expected) -> None:
    from dioptra_builtins.metrics.performance import categorical_accuracy

    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    result: float = categorical_accuracy(y_true, y_pred)
    assert np.isclose(expected, result)


@pytest.mark.parametrize(
    ("y_true", "y_pred", "weights", "expected"),
    [
        ([0, 1], [1, 1], None, 0.0),
        ([1, 2, 4], [4, 1, 1], [0.3, 0.6, 0.1], -0.5039526306789696),
        ([1, 2, 3, 6, 5, 6], [6, 6, 6, 6, 6, 6], None, 0.0),
        ([9, 9, 9, 9, 9, 2], [9, 9, 9, 9, 9, 9], [0.2, 0.1, 0.3, 0.1, 0.1, 0.2], 0.0),
        (
            [2, 2, 3, 4, 5],
            [2, 2, 5, 5, 5],
            [0.2, 0.2, 0.2, 0.2, 0.2],
            0.5443310539518176,
        ),
    ],
)
def test_mcc(y_true, y_pred, weights, expected) -> None:
    from dioptra_builtins.metrics.performance import mcc

    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    kwargs = {"sample_weight": weights}
    result: float = mcc(y_true, y_pred, **kwargs)
    assert np.isclose(expected, result)


@pytest.mark.parametrize(
    ("y_true", "y_pred", "weights", "average", "expected"),
    [
        ([0, 1], [1, 1], None, "micro", 0.5),
        ([1, 2, 4], [4, 1, 1], [0.3, 0.6, 0.1], "macro", 0.0),
        ([1, 2, 3, 6, 5, 6], [6, 6, 6, 6, 6, 6], None, "weighted", 0.16666666),
        (
            [9, 9, 9, 9, 9, 2],
            [9, 9, 9, 9, 9, 9],
            [0.2, 0.1, 0.3, 0.1, 0.1, 0.2],
            "micro",
            0.8,
        ),
        ([2, 2, 3, 4, 5], [2, 2, 5, 5, 5], [0.2, 0.2, 0.2, 0.2, 0.2], "weighted", 0.5),
    ],
)
def test_f1(y_true, y_pred, weights, average, expected) -> None:
    from dioptra_builtins.metrics.performance import f1

    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    kwargs = {"average": average, "sample_weight": weights}
    result: float = f1(y_true, y_pred, **kwargs)
    assert np.isclose(result, expected)


@pytest.mark.parametrize(
    ("y_true", "y_pred", "weights", "average", "expected"),
    [
        ([0, 1], [1, 1], None, "micro", 0.5),
        ([1, 2, 4], [4, 1, 1], [0.3, 0.6, 0.1], "macro", 0.0),
        ([1, 2, 3, 6, 5, 6], [6, 6, 6, 6, 6, 6], None, "weighted", 0.1111111),
        (
            [9, 9, 9, 9, 9, 2],
            [9, 9, 9, 9, 9, 9],
            [0.2, 0.1, 0.3, 0.1, 0.1, 0.2],
            "micro",
            0.8,
        ),
        (
            [2, 2, 3, 4, 5],
            [2, 2, 5, 5, 5],
            [0.2, 0.2, 0.2, 0.2, 0.2],
            "weighted",
            0.466666666,
        ),
    ],
)
def test_precision(y_true, y_pred, weights, average, expected) -> None:
    from dioptra_builtins.metrics.performance import precision

    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    kwargs = {"average": average, "sample_weight": weights}
    result: float = precision(y_true, y_pred, **kwargs)
    assert np.isclose(result, expected)


@pytest.mark.parametrize(
    ("y_true", "y_pred", "weights", "average", "expected"),
    [
        ([0, 1], [1, 1], None, "micro", 0.5),
        ([1, 2, 4], [4, 1, 1], [0.3, 0.6, 0.1], "macro", 0.0),
        ([1, 2, 3, 6, 5, 6], [6, 6, 6, 6, 6, 6], None, "weighted", 0.3333333),
        (
            [9, 9, 9, 9, 9, 2],
            [9, 9, 9, 9, 9, 9],
            [0.2, 0.1, 0.3, 0.1, 0.1, 0.2],
            "micro",
            0.8,
        ),
        ([2, 2, 3, 4, 5], [2, 2, 5, 5, 5], [0.2, 0.2, 0.2, 0.2, 0.2], "weighted", 0.6),
    ],
)
def test_recall(y_true, y_pred, weights, average, expected) -> None:
    from dioptra_builtins.metrics.performance import recall

    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    kwargs = {"average": average, "sample_weight": weights}
    result: float = recall(y_true, y_pred, **kwargs)
    assert np.isclose(result, expected)



"""empty message

Revision ID: 018130a0bf6c
Revises: 32d5c0055e1b
Create Date: 2022-08-24 18:34:48.570620

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "018130a0bf6c"
down_revision = "32d5c0055e1b"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "users",
        sa.Column(
            "user_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("username", sa.Text(), nullable=False),
        sa.Column("password", sa.Text(), nullable=False),
        sa.Column("email_address", sa.Text(), nullable=False),
        sa.Column("created_on", sa.DateTime(), nullable=False),
        sa.Column("last_modified_on", sa.DateTime(), nullable=False),
        sa.Column("last_login_on", sa.DateTime(), nullable=False),
        sa.Column("user_expire_on", sa.DateTime(), nullable=False),
        sa.Column("password_expire_on", sa.DateTime(), nullable=False),
        sa.Column("is_deleted", sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint("user_id", name=op.f("pk_users")),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("users")
    # ### end Alembic commands ###
